#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     29.08.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#

from dataclasses import dataclass
import statistics
import math
from typing import Optional

from sniffles import sv
from sniffles import leadprov


@dataclass
class Cluster:
    id: str
    svtype: str
    contig: str
    start: int
    end: int
    seed: int
    leads: list
    repeat: bool
    leads_long: Optional[list]

    @property
    def span(self) -> Optional[int]:
        if self.end is None or self.start is None:
            return None

        return self.end - self.start

    def compute_metrics(self, max_n=100):
        n = min(len(self.leads), max_n)
        if n == 0:
            self.mean_svlen = 0
            self.stdev_start = 0
            return

        step = int(len(self.leads) / n)
        if n > 1:
            self.mean_svlen = sum(self.leads[i].svlen for i in range(0, len(self.leads), step)) / float(n)
            self.stdev_start = statistics.stdev(self.leads[i].ref_start for i in range(0, len(self.leads), step))
        else:
            self.mean_svlen = self.leads[0].svlen
            self.stdev_start = 0


def merge_inner(cluster, threshold):
    read_seq = {}
    for ld in cluster.leads:
        if ld.read_qname not in read_seq:
            read_seq[ld.read_qname] = []
        read_seq[ld.read_qname].append(ld)

    cluster.leads = []
    for qname in read_seq:
        read_seq[qname].sort(key=lambda k: k.ref_start)
        to_merge = read_seq[qname][0]

        curr_lead = to_merge

        last_ref_end = to_merge.ref_end
        last_qry_end = to_merge.qry_end
        last_ref_start = to_merge.ref_start
        last_qry_start = to_merge.qry_start

        for to_merge in read_seq[qname][1:]:
            merge = (threshold == -1) or ((abs(to_merge.ref_start - last_ref_end) < threshold or abs(to_merge.ref_start - last_ref_start) < threshold) and (
                        abs(to_merge.qry_start - last_qry_end) < threshold or abs(to_merge.qry_start - last_qry_start) < threshold))
            if merge:
                curr_lead.svlen += to_merge.svlen
                if to_merge.seq is None or curr_lead.seq is None:
                    curr_lead.seq = None
                else:
                    curr_lead.seq += to_merge.seq
            else:
                cluster.leads.append(curr_lead)
                curr_lead = to_merge
            last_ref_end = to_merge.ref_end
            last_qry_end = to_merge.qry_end
            last_ref_start = to_merge.ref_start
            last_qry_start = to_merge.qry_start

        cluster.leads.append(curr_lead)
    return cluster


def resplit(cluster, prop, binsize, merge_threshold_min, merge_threshold_frac):
    bins_leads = {}
    for lead in cluster.leads:
        bin = int(abs(prop(lead)) / binsize) * binsize
        if not bin in bins_leads:
            bins_leads[bin] = [lead]
        else:
            bins_leads[bin].append(lead)

    new_clusters = list(sorted(bins_leads.keys()))
    i = 1
    while len(new_clusters) > 1 and i < len(new_clusters):
        last_cluster = new_clusters[i - 1]
        curr_cluster = new_clusters[i]
        merge_threshold = max(merge_threshold_min, min(curr_cluster, last_cluster) * merge_threshold_frac)
        merge = abs(curr_cluster - last_cluster) <= merge_threshold
        if merge:
            bins_leads[new_clusters[i]].extend(bins_leads[new_clusters[i - 1]])
            new_clusters.pop(i - 1)
            i = max(0, i - 2)
        else:
            i += 1

    for cluster_index in new_clusters:
        new_cluster = Cluster(id=cluster.id + f".{cluster_index}",
                              svtype=cluster.svtype,
                              contig=cluster.contig,
                              start=cluster.start,
                              end=cluster.end,
                              seed=cluster.seed,
                              leads=bins_leads[cluster_index],
                              repeat=cluster.repeat,
                              leads_long=cluster.leads_long)
        yield new_cluster


def resplit_bnd(cluster, merge_threshold):
    contigs_leads = {}

    # Split by mate contig and mate ref start
    for lead in cluster.leads:
        if not lead.bnd_info.mate_contig in contigs_leads:
            contigs_leads[lead.bnd_info.mate_contig] = {}

        bin = int(lead.bnd_info.mate_ref_start / merge_threshold) * merge_threshold if merge_threshold > 0 else 0
        if not bin in contigs_leads[lead.bnd_info.mate_contig]:
            contigs_leads[lead.bnd_info.mate_contig][bin] = [lead]
        else:
            contigs_leads[lead.bnd_info.mate_contig][bin].append(lead)

    for contig in contigs_leads:
        bins = sorted(contigs_leads[contig])
        curr_leads = [] + contigs_leads[contig][bins[0]]
        last_bin = bins[0]
        for bin in bins[1:]:
            if bin - last_bin <= merge_threshold:
                curr_leads.extend(contigs_leads[contig][bin])
            else:
                if len(curr_leads):
                    new_cluster = Cluster(id=cluster.id + f".CHR2.{contig}.POS2.{bin}",
                                          svtype=cluster.svtype,
                                          contig=cluster.contig,
                                          start=cluster.start,
                                          end=cluster.end,
                                          seed=cluster.seed,
                                          leads=[k for k in curr_leads],
                                          repeat=cluster.repeat,
                                          leads_long=None)
                    yield new_cluster
                curr_leads = [] + contigs_leads[contig][bin]
            last_bin = bin
        if len(curr_leads):
            new_cluster = Cluster(id=cluster.id + f".CHR2.{contig}.POS2.{bin}",
                                  svtype=cluster.svtype,
                                  contig=cluster.contig,
                                  start=cluster.start,
                                  end=cluster.end,
                                  seed=cluster.seed,
                                  leads=[k for k in curr_leads],
                                  repeat=cluster.repeat,
                                  leads_long=None)
            curr_leads = []
            yield new_cluster


def resolve(svtype, leadtab_provider, config, tr):
    leadtab = leadtab_provider.leadtab[svtype]
    seeds = sorted(leadtab_provider.leadtab[svtype])

    if len(seeds) == 0:
        return []

    # Initialize tandem repeat region handling
    if tr is not None:
        tr_index = 0
        if len(tr) == 0:
            tr = None
        else:
            tr_start, tr_end = tr[tr_index]

    clusters = []
    for seed_index, seed in enumerate(seeds):
        if config.dev_call_region != None:
            if seed < config.dev_call_region["start"] or seed > config.dev_call_region["end"]:
                continue

        within_tr = False
        if tr is not None and tr_index < len(tr):
            while tr_end < seed and tr_index + 1 < len(tr):
                tr_index += 1
                tr_start, tr_end = tr[tr_index]
            if tr_start < seed < tr_end:
                within_tr = True

        if svtype == "INS":
            leads = [lead for lead in leadtab[seed] if lead.svlen != None]
            leads_long = [lead for lead in leadtab[seed] if lead.svlen == None]
        else:
            leads = leadtab[seed]
            leads_long = None

        cluster = Cluster(id=f"CL.{svtype}.{leadtab_provider.contig}.{leadtab_provider.start}.{seed_index}",
                          svtype=svtype,
                          contig=leadtab_provider.contig,
                          start=seed,
                          end=seed + config.cluster_binsize,
                          seed=seed,
                          leads=leads,
                          repeat=within_tr or config.repeat,
                          leads_long=leads_long)

        cluster.compute_metrics()
        clusters.append(cluster)

    # Merge clusters
    cluster_count_initial = len(clusters)
    i = 0
    while i < len(clusters) - 1:
        curr_cluster = clusters[i]
        next_cluster = clusters[i + 1]

        # assert((next_cluster.end - curr_cluster.start) >= 0)
        # assert((next_cluster.start - curr_cluster.end) >= 0)

        # Fast Adaptive Scanning Clustering
        #  Inter-Cluster distance < (StdDev-Start-Within-A + StdDev-Stat-Within-B) * r   (r=1.0 default)
        #  Merged size (OR: Inter-cluster distance) < (MeanSVlen between both clusters) * h (h=0.5 default)
        #  Max. overall cluster size criterion?

        inner_dist = (next_cluster.start - curr_cluster.end)
        outer_dist = (next_cluster.end - curr_cluster.start)
        merge = inner_dist <= min(curr_cluster.stdev_start, next_cluster.stdev_start) * config.cluster_r
        merge = merge or ((config.repeat or curr_cluster.repeat or next_cluster.repeat) and outer_dist <= min(config.cluster_repeat_h_max,
                                                                                                              (abs(curr_cluster.mean_svlen) + abs(next_cluster.mean_svlen)) * config.cluster_repeat_h))
        merge = merge or (svtype == "BND" and inner_dist <= config.cluster_merge_bnd)

        if merge:
            clusters.pop(i + 1)
            curr_cluster.leads += next_cluster.leads
            if svtype == "INS":
                curr_cluster.leads_long += next_cluster.leads_long
            curr_cluster.end = next_cluster.end
            curr_cluster.repeat = curr_cluster.repeat or next_cluster.repeat
            curr_cluster.compute_metrics()
            i = max(0, i - 2)
        i += 1

    if config.dev_trace_read:
        for c in clusters:
            for ld in c.leads:
                if ld.read_qname == config.dev_trace_read:
                    print(f"[DEV_TRACE_READ [2/4] [cluster.resolve] Read lead {ld} is in cluster {c.id}, containing a total of {len(c.leads)} leads")

    if config.dev_dump_clusters:
        filename = f"{config.input}.clusters.{svtype}.{leadtab_provider.contig}.{leadtab_provider.start}.{leadtab_provider.end}.bed"
        print(f"Dumping clusters to {filename}")
        with open(filename, "w") as h:
            for c in clusters:
                info = f"ID={c.id}, #LEADS={len(c.leads)}; "
                for ld in c.leads:
                    info += f"(ref_start={ld.ref_start},svlen={ld.svlen},source={ld.source}); "
                h.write(f"{c.contig}\t{c.start}\t{c.end}\t\"{info}\"\n")

    for cluster in clusters:
        if len(cluster.leads) == 0:
            continue

        if svtype == "BND":
            if config.dev_no_resplit:
                yield cluster
            else:
                for new_cluster in resplit_bnd(cluster, merge_threshold=config.cluster_merge_bnd):
                    yield new_cluster
        else:
            if svtype == "INS" or svtype == "DEL":
                if cluster.repeat:
                    merge_inner_threshold = -1
                else:
                    merge_inner_threshold = config.cluster_merge_pos

                merge_inner(cluster, merge_inner_threshold)

            if not config.dev_no_resplit_repeat and not config.dev_no_resplit:
                for new_cluster in resplit(cluster,
                                           prop=lambda lead: lead.svlen,
                                           binsize=config.cluster_resplit_binsize,
                                           merge_threshold_min=config.minsvlen,
                                           merge_threshold_frac=config.cluster_merge_len):
                    yield new_cluster
            else:
                yield cluster


def resolve_block_groups(svtype, svcands, groups_initial, config):
    """For clustering groups of SVs for combining .snfs (multi-call)"""

    # TODO: Remove sorting
    groups = groups_initial
    for svcand in sorted(svcands, key=lambda cand: cand.support, reverse=True):
        best_group = None
        best_dist = math.inf

        if svtype == "BND":
            mate_contig, mate_ref_start = svcand.bnd_info.mate_contig, svcand.bnd_info.mate_ref_start
            for group in groups:
                # TODO: Favor bigger groups in placement
                dist = abs(group.pos_mean - svcand.pos) + abs(group.bnd_mate_ref_start_mean - mate_ref_start)
                if dist < best_dist and dist <= config.cluster_merge_bnd * 2 and group.bnd_mate_contig == mate_contig:
                    if not config.combine_separate_intra or svcand.sample_internal_id not in group.included_samples:
                        best_group = group
                        best_dist = dist
        else:
            for group in groups:
                # TODO: Favor bigger groups in placement
                dist = abs(group.pos_mean - svcand.pos) + abs(abs(group.len_mean) - abs(svcand.svlen))  # check if group.pos_mean is updated or stays the same for the first SV starting the group
                minlen = float(min(abs(group.len_mean), abs(svcand.svlen)))
                if minlen > 0 and dist < best_dist and dist <= config.combine_match * math.sqrt(minlen) and dist <= config.combine_match_max:
                    if (not config.combine_separate_intra or svcand.sample_internal_id not in group.included_samples) and group.align_call(svcand, config.combine_pctseq):
                        best_group = group
                        best_dist = dist

        if best_group is None:
            groups.append(
                sv.SVGroup.from_candidate(svcand)
            )
        else:
            best_group.add_candidate(svcand)
    return groups



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     15.08.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
import logging
from collections import Counter

import pysam
import os

from sniffles import sv
from sniffles import util
from sniffles.config import SnifflesConfig


log = logging.getLogger(__name__)


def format_info(k, v):
    if isinstance(v, float):
        return f"{k}={v:.3f}"
    elif isinstance(v, list):
        return f"{k}={','.join(v)}"
    elif v is None:
        v = '.'

    return f"{k}={v}"


def unpack_phase(phase, svid="") -> tuple:
    try:
        hp_i, ps = phase
    except TypeError:
        if phase is None:
            hp_i, ps = ".", "."
        else:
            log.debug(f"Single not 'None'-valued phase: {phase}|{svid}")
            hp_i, ps = phase, phase
    ps = ps if ps is not None else "."
    return hp_i, ps


def format_genotype(gt, is_phased):
    """
    hp_i is the index of the haplotype in config.phase_identifiers:
    HP:1 => index 0 => phased genotype in the form of 1|0
    HP:2 => index 1 => phased genotype in the form of 0|1
    is_phased was added to modify the output as phasing adds the PS tag to FORMAT
    """
    gt_multi_sample_fields = 6
    if len(gt) == gt_multi_sample_fields:
        a, b, qual, dr, dv, phase = gt
        hp_i, ps = unpack_phase(phase)
        if hp_i is not None and (a, b) == (0, 1) and is_phased:
            gt_sep = "|"
            if hp_i == 0:
                a, b = b, a
        else:
            gt_sep = "/"
        return f"{a}{gt_sep}{b}:{qual}:{dr}:{dv}:{ps}" if is_phased else f"{a}{gt_sep}{b}:{qual}:{dr}:{dv}"
    else:
        a, b, qual, dr, dv, phase, svid = gt
        hp_i, ps = unpack_phase(phase, svid)
        if hp_i is not None and (a, b) == (0, 1) and is_phased:
            gt_sep = "|"
            if hp_i == 0:
                a, b = b, a
        else:
            gt_sep = "/"
        return f"{a}{gt_sep}{b}:{qual}:{dr}:{dv}:{ps}:{svid}" if is_phased \
            else f"{a}{gt_sep}{b}:{qual}:{dr}:{dv}:{svid}"


class VCF:
    def __init__(self, config: SnifflesConfig, handle):
        self.config = config
        self.handle = handle
        self.call_count = 0
        self.info_order = ["SVTYPE", "SVLEN", "END", "SUPPORT", "RNAMES", "COVERAGE", "STRAND"]
        if config.qc_nm_measure:
            self.info_order.append("NM")

        if config.dev_emit_sv_lengths:
            self.info_order.append("SVLENGTHS")

        self.default_genotype = config.genotype_none

        # Add phasing if needed
        self.genotype_format = config.genotype_format
        if config.phase:
            self.genotype_format += ":PS"
            # it has it added already as None
        if config.mode == "combine":
            self.genotype_format += ":ID"
            self.default_genotype += tuple(["NULL"])

        self.reference_handle = None
        self.header_str = ""

    def open_reference(self, generate_index: bool = True):
        if self.config.reference is None:
            return

        if not os.path.exists(self.config.reference + ".fai") and not os.path.exists(self.config.reference + ".gzi") and generate_index:
            log.warning(f"Fasta index for {self.config.reference} not found. Generating with pysam.faidx (this may take a while)")
            pysam.faidx(self.config.reference)

        try:
            self.reference_handle = pysam.FastaFile(self.config.reference)
        except:
            log.error(f'Unable to open reference file {self.config.reference}')

    def write_header(self, contigs_lengths):
        self.write_header_line("fileformat=VCFv4.2")
        self.write_header_line(f"source={self.config.version}_{self.config.build}")
        self.write_header_line('command="' + self.config.command + '"')
        self.write_header_line('fileDate="' + self.config.start_date + '"')
        for contig, contig_len in contigs_lengths:
            self.write_header_line(f"contig=<ID={contig},length={contig_len}>")

        self.write_header_line('ALT=<ID=INS,Description="Insertion">')
        self.write_header_line('ALT=<ID=DEL,Description="Deletion">')
        self.write_header_line('ALT=<ID=DUP,Description="Duplication">')
        self.write_header_line('ALT=<ID=INV,Description="Inversion">')
        self.write_header_line('ALT=<ID=BND,Description="Breakend; Translocation">')

        self.write_header_line('FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">')
        self.write_header_line('FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype quality">')
        self.write_header_line('FORMAT=<ID=DR,Number=1,Type=Integer,Description="Number of reference reads">')
        self.write_header_line('FORMAT=<ID=DV,Number=1,Type=Integer,Description="Number of variant reads">')
        self.write_header_line('FORMAT=<ID=PS,Number=1,Type=Integer,Description="Phase-block, zero if none or not phased">')
        self.write_header_line('FORMAT=<ID=ID,Number=1,Type=String,Description="Individual sample SV ID for multi-sample output">')

        self.write_header_line('FILTER=<ID=PASS,Description="All filters passed">')
        self.write_header_line('FILTER=<ID=GT,Description="Genotype filter">')
        self.write_header_line('FILTER=<ID=SUPPORT_MIN,Description="Minimum read support filter">')
        self.write_header_line('FILTER=<ID=STDEV_POS,Description="SV Breakpoint standard deviation filter">')
        self.write_header_line('FILTER=<ID=STDEV_LEN,Description="SV length standard deviation filter">')
        self.write_header_line('FILTER=<ID=COV_MIN,Description="Minimum coverage filter">')
        self.write_header_line('FILTER=<ID=COV_MIN_GT,Description="Minimum coverage filter (missing genotype)">')
        self.write_header_line('FILTER=<ID=COV_CHANGE_DEL,Description="Coverage change filter for DEL">')
        self.write_header_line('FILTER=<ID=COV_CHANGE_DUP,Description="Coverage change filter for DUP">')
        self.write_header_line('FILTER=<ID=COV_CHANGE_INS,Description="Coverage change filter for INS">')
        self.write_header_line('FILTER=<ID=COV_CHANGE_FRAC_US,Description="Coverage fractional change filter: upstream-start">')
        self.write_header_line('FILTER=<ID=COV_CHANGE_FRAC_SC,Description="Coverage fractional change filter: start-center">')
        self.write_header_line('FILTER=<ID=COV_CHANGE_FRAC_CE,Description="Coverage fractional change filter: center-end">')
        self.write_header_line('FILTER=<ID=COV_CHANGE_FRAC_ED,Description="Coverage fractional change filter: end-downstream">')
        self.write_header_line('FILTER=<ID=MOSAIC_VAF,Description="Mosaic variant allele fraction filter">')
        self.write_header_line('FILTER=<ID=NOT_MOSAIC_VAF,Description="Variant allele fraction filter for non-mosaic">')
        self.write_header_line('FILTER=<ID=ALN_NM,Description="Length adjusted mismatch filter">')
        self.write_header_line('FILTER=<ID=STRAND_BND,Description="Strand support filter for BNDs">')
        self.write_header_line('FILTER=<ID=STRAND,Description="Strand support filter for germline SVs">')
        self.write_header_line('FILTER=<ID=STRAND_MOSAIC,Description="Strand support filter for mosaic SVs">')
        self.write_header_line('FILTER=<ID=SVLEN_MIN,Description="SV length filter">')
        self.write_header_line('FILTER=<ID=SVLEN_MIN_MOSAIC,Description="SV length filter for mosaic SVs">')
        self.write_header_line('INFO=<ID=PRECISE,Number=0,Type=Flag,Description="Structural variation with precise breakpoints">')
        self.write_header_line('INFO=<ID=IMPRECISE,Number=0,Type=Flag,Description="Structural variation with imprecise breakpoints">')
        self.write_header_line('INFO=<ID=MOSAIC,Number=0,Type=Flag,Description="Structural variation classified as putative mosaic">')
        self.write_header_line('INFO=<ID=SVLEN,Number=1,Type=Integer,Description="Length of structural variation">')
        if self.config.dev_emit_sv_lengths:
            self.write_header_line('INFO=<ID=SVLENGTHS,Number=.,Type=Integer,Description="Lengths of structural variation (all)">')
        self.write_header_line('INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variation">')
        self.write_header_line('INFO=<ID=CHR2,Number=1,Type=String,Description="Mate chromsome for BND SVs">')
        self.write_header_line('INFO=<ID=SUPPORT,Number=1,Type=Integer,Description="Number of reads supporting the structural variation">')
        self.write_header_line('INFO=<ID=SUPPORT_INLINE,Number=1,Type=Integer,Description="Number of reads supporting an INS/DEL SV (non-split events only)">')
        self.write_header_line('INFO=<ID=SUPPORT_SA,Number=1,Type=Integer,Description="Number of reads supporting a DEL SV through supplementary alignments (split events)">')
        self.write_header_line('INFO=<ID=SUPPORT_LONG,Number=1,Type=Integer,Description="Number of soft-clipped reads putatively supporting the long insertion SV">')
        self.write_header_line('INFO=<ID=END,Number=1,Type=Integer,Description="End position of structural variation">')
        self.write_header_line('INFO=<ID=STDEV_POS,Number=1,Type=Float,Description="Standard deviation of structural variation start position">')
        self.write_header_line('INFO=<ID=STDEV_LEN,Number=1,Type=Float,Description="Standard deviation of structural variation length">')
        self.write_header_line('INFO=<ID=COVERAGE,Number=.,Type=Float,Description="Coverages near upstream, start, center, end, downstream of structural variation">')
        self.write_header_line('INFO=<ID=STRAND,Number=1,Type=String,Description="Strands of supporting reads for structural variant">')
        self.write_header_line('INFO=<ID=AC,Number=.,Type=Integer,Description="Allele count, summed up over all samples">')
        self.write_header_line('INFO=<ID=SUPP_VEC,Number=1,Type=String,Description="List of read support for all samples">')
        self.write_header_line('INFO=<ID=CONSENSUS_SUPPORT,Number=1,Type=Integer,Description="Number of reads that support the generated insertion (INS) consensus sequence">')
        self.write_header_line('INFO=<ID=RNAMES,Number=.,Type=String,Description="Names of supporting reads (if enabled with --output-rnames)">')
        self.write_header_line('INFO=<ID=VAF,Number=1,Type=Float,Description="Variant Allele Fraction">')
        self.write_header_line('INFO=<ID=NM,Number=.,Type=Float,Description="Mean number of query alignment length adjusted mismatches of supporting reads">')
        self.write_header_line('INFO=<ID=PHASE,Number=.,Type=String,Description="Phasing information derived from supporting reads, represented as list of: HAPLOTYPE,PHASESET,HAPLOTYPE_SUPPORT,PHASESET_SUPPORT,HAPLOTYPE_FILTER,PHASESET_FILTER">')

        if self.config.combine_population:
            self.write_header_line('INFO=<ID=POPULATION_VAF,Number=1,Type=Float,Description="Variant Allele Fraction in population">')
            self.write_header_line('INFO=<ID=POPULATION_SIZE,Number=1,Type=Integer,Description="Size of genotyped population for this variant">')

        samples_header = "\t".join(sample_id for _, sample_id in self.config.sample_ids_vcf)
        self.write_raw(f"#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t{samples_header}")

    def write_raw(self, text, endl="\n"):
        if self.config.vcf_output_bgz:
            self.handle.write(text.encode())
            self.handle.write(endl.encode())
        else:
            self.handle.write(text)
            self.handle.write(endl)

    def write_header_line(self, text):
        self.write_raw("##" + text)

    def write_call(self, call) -> int:
        # pysam coordinates are 0-based, VCF 1-based
        # but VCF also requires the index of the base before the SV to be reported,
        # so we are fine without offsetting
        end = call.end
        pos = call.pos if call.pos > 0 else 1

        # Determine genotypes columns
        ac = 0  # Allele count
        supvec = []
        sample_genotypes = []
        for internal_id, _ in self.config.sample_ids_vcf:
            if internal_id in call.genotypes and call.genotypes[internal_id] is not None:
                gt_curr = call.genotypes[internal_id]
                sample_genotypes.append(format_genotype(gt_curr, self.config.phase))
                if gt_curr[0] != "." and gt_curr[4] > 0:  # Not non-genotype and has supporting reads
                    ac += sum(call.genotypes[internal_id][:2])
                    supp = "1"
                else:
                    supp = "0"
            else:
                sample_genotypes.append(format_genotype(self.default_genotype, self.config.phase))
                supp = "0"
            supvec.append(supp)

        if len(self.config.sample_ids_vcf) > 1:
            call.set_info("AC", ac)
            call.set_info("SUPP_VEC", svec := "".join(supvec))

            if int(svec) == 0:
                log.debug(f'Dropped {call} due to all zero support vector.')
                return 0

            if ac == 0:
                call.filter = "GT"

        # Check if svlen == len(alt) in INS
        if "INS" == call.svtype:
            if call.svlen != (len(call.alt) - 1) and not self.config.symbolic:
                log.debug(f"Updating SVLEN for INS to match sequence length: {call.svlen} v {(len(call.alt) - 1)}")
                call.svlen = (len(call.alt) - 1)

        # Output core SV attributes
        infos = {
            "SVTYPE": call.svtype,
            "SVLEN": call.svlen,
            "SVLENGTHS": ",".join(map(str, call.svlens)) if call.svlens else None,
            "END": end,
            "SUPPORT": call.support,
            "RNAMES": call.rnames if self.config.output_rnames else None,
            "COVERAGE": f"{call.coverage_upstream},{call.coverage_start},{call.coverage_center},{call.coverage_end},"
                        f"{call.coverage_downstream}",
            "STRAND": ("+" if call.fwd > 0 else "") + ("-" if call.rev > 0 else ""),
            "NM": call.nm
        }

        if call.svtype == "BND":
            infos["SVLEN"] = None
            infos["SVLENGTHS"] = None
            infos["END"] = None

        if call.svtype == "DEL":
            # END is POS + length of REF allele - 1
            infos["END"] = end - 1

        infos_ordered = ["PRECISE" if call.precise else "IMPRECISE"]
        af = call.get_info("VAF")
        af = af if af is not None else 0
        sv_is_mosaic = af <= self.config.mosaic_af_max
        if sv_is_mosaic and self.config.mosaic:
            infos_ordered.append("MOSAIC")
        infos_ordered.extend(format_info(k, infos[k]) for k in self.info_order if infos[k] is not None)
        info_str = ";".join(infos_ordered)

        # Output call specific additional information
        for k in sorted(call.info):
            if call.info[k] is None:
                continue
            info_str += ";" + format_info(k, call.info[k])

        # if call.id==None:
        #    call.id=f"Sniffles2.{call.svtype}.{self.call_count+1:06}"

        # Resolve DEL sequence
        if (not self.config.symbolic and call.svtype == "DEL" and self.reference_handle is not None
                and abs(call.svlen) <= self.config.max_del_seq_len):
            try:
                # VCF requires inclusion of the last reference base before the SV
                call.ref = self.reference_handle.fetch(call.contig, call.pos - 1, call.pos - call.svlen)
                call.alt = call.ref[0]
            except KeyError:
                call.ref = "N"
                call.alt = f"<{call.svtype}>"
            except ValueError:
                call.ref = "N"
                call.alt = f"<{call.svtype}>"
            else:
                if 'N' in call.ref and (pct_n := Counter(call.ref)['N'] / len(call.ref)) > self.config.max_unknown_pct:
                    # don't emit calls with too many N bases
                    log.debug(f'Not emitting {call.id} (length {call.svlen}) due to {pct_n*100:.2f}% N bases in reference.')
                    return 0

        if self.config.symbolic:
            call.ref = "N"
            call.alt = f"<{call.svtype}>"
        else:
            if self.reference_handle is not None and call.ref == 'N':
                # Fetch the base before the SV
                try:
                    call.ref = self.reference_handle.fetch(call.contig, start := max(0, call.pos - 1), start + 1)
                except (KeyError, ValueError):
                    ...
                else:
                    if call.svtype == "INS" and call.alt != '<INS>':
                        call.alt = call.ref + call.alt
                    elif call.svtype == 'BND' and call.alt != '<BND>':
                        call.alt = (call.ref + call.alt[1:]) if call.alt.startswith('N') else call.alt[:-1] + call.ref

        call.qual = max(0, min(60, call.qual)) if call.qual is not None else None

        self.write_raw("\t".join(str(v) for v in [call.contig, pos, self.config.id_prefix + call.id, call.ref,
                                                  call.alt, call.qual if call.qual is not None else '.', call.filter, info_str, self.genotype_format] +
                                 sample_genotypes))
        self.call_count += 1
        return 1

    def read_svs_iter(self):
        self.header_str = ""
        line_index = 0
        for line in self.handle:
            try:
                if isinstance(line, bytes):
                    line = line.decode("utf-8")
                line_index += 1
                line_strip = line.strip()
                if line_strip == "" or line_strip[0] == "#":
                    if line_strip[0] == "#":
                        self.header_str += line_strip + "\n"
                    continue
                CHROM, POS, _, REF, ALT, QUAL, FILTER, INFO = line.split("\t")[:8]
                info_dict = {}
                for info_item in INFO.split(";"):
                    if "=" in info_item:
                        key, value = info_item.split("=")
                    else:
                        key, value = info_item, True
                    info_dict[key] = value
                call = sv.SVCall(contig=CHROM,
                                 pos=int(POS) - 1,
                                 id=line_index,
                                 ref=REF,
                                 alt=ALT,
                                 qual=int(QUAL) if QUAL != '.' else None,
                                 filter=FILTER,
                                 info=info_dict,
                                 svtype=None,
                                 svlen=None,
                                 end=None,
                                 rnames=None,
                                 qc=True,
                                 postprocess=None,
                                 genotypes=None,
                                 precise=None,
                                 support=0,
                                 fwd=0,
                                 rev=0,
                                 nm=-1)
                if len(call.alt) > len(call.ref):
                    call.svtype = "INS"
                    call.svlen = len(call.alt)
                    call.end = call.pos
                else:
                    call.svtype = "DEL"
                    call.svlen = -len(call.ref)
                    call.end = call.pos + call.svlen

                if "SVTYPE" in info_dict:
                    call.svtype = info_dict["SVTYPE"]
                    if call.svtype == "TRA":
                        call.svtype = "BND"

                if "SVLEN" in info_dict:
                    call.svlen = int(info_dict["SVLEN"])
                if "SVLENGTHS`" in info_dict:
                    call.svlens = info_dict["SVLENGTHS"]

                if "END" in info_dict:
                    call.end = int(info_dict["END"])

                if call.svtype == "BND":
                    bnd_parts = call.alt.replace("]", "[").split("[")
                    if len(bnd_parts) > 2:
                        mate_contig, mate_ref_start = bnd_parts[1].split(":")
                        call.bnd_info = sv.SVCallBNDInfo(mate_contig=mate_contig, mate_ref_start=int(mate_ref_start),
                                                         is_first=(call.alt[0] == "N"), is_reverse=("]" in call.alt))
                    else:
                        raise ValueError("BND ALT not formatted according to VCF 4.2 specifications")

                call.raw_vcf_line = line_strip
                call.raw_vcf_line_index = line_index
                yield call
            except Exception as e:
                util.fatal_error(f"Error parsing input VCF: Line {line_index}: {e}")

    def rewrite_genotype(self, svcall):
        parts_no_gt = svcall.raw_vcf_line.split("\t")[:8]
        gt_format = self.config.genotype_format
        if svcall.genotype_match_sv != None:
            if len(svcall.genotype_match_sv.genotypes) > 0:
                gt = svcall.genotype_match_sv.genotypes[0]
            else:
                gt = svcall.genotypes[0]
        else:
            gt = svcall.genotypes[0]  # 0/0 or ./. depending on whether input SV had coverage in sample
        # parts=parts_no_gt + [gt_format,vcf.format_genotype(gt)]
        # gt_vcf=svcall.raw_vcf_line.split("\t")[9].split(":")[0]
        # parts= parts_no_gt + [gt_vcf] + [gt_format,vcf.format_genotype(gt)]
        parts = parts_no_gt + [gt_format, format_genotype(gt, self.config.phased)]
        # parts[7]="NA"
        # parts[3]=f"REF_{len(parts[3])}"
        # parts[4]=f"ALT_{len(parts[4])}"
        self.write_raw("\t".join(parts))

    def rewrite_header_genotype(self,orig_header):
        header_lines=orig_header.split("\n")
        header_lines.insert(1,'##genotypeFileDate="'+self.config.start_date+'"')
        header_lines.insert(1,'##genotypeCommand="'+self.config.command+'"')
        header_lines.insert(1,f"##genotypeSource={self.config.version}_{self.config.build}")

        # Fix missing genotype headers errors when reading input vcf in genotype mode.
        # This error will happen when input vcf does not have GT,GQ,DR,DV headers.
        # Some tools such as turvari will throw errors when processing the output vcf.
        has_gt_headers = {
            "GT": False,
            "GQ": False,
            "DR": False,
            "DV": False,
        }
        for header_line in header_lines:
            for gt in has_gt_headers.keys():
                if "##FORMAT=<ID="+gt+"," in header_line:
                    has_gt_headers[gt] = True
        
        if not has_gt_headers["GT"]:
            header_lines.insert(len(header_lines)-2, '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">')
        if not has_gt_headers["GQ"]:
            header_lines.insert(len(header_lines)-2, '##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype quality">')
        if not has_gt_headers["DR"]:
            header_lines.insert(len(header_lines)-2, '##FORMAT=<ID=DR,Number=1,Type=Integer,Description="Number of reference reads">')
        if not has_gt_headers["DV"]:
            header_lines.insert(len(header_lines)-2, '##FORMAT=<ID=DV,Number=1,Type=Integer,Description="Number of variant reads">')

        self.write_raw("\n".join(header_lines), endl="")

    def close(self):
        self.handle.close()



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     27.08.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#

import statistics
import sys
import time


class Sniffles2Exit(Exception):
    code: int = 1


def stdev(nums):
    nums = list(nums)
    return statistics.stdev(nums) if len(nums) > 1 else 0


def median(nums):
    return int(statistics.median(nums))


def median_or_mode(nums):
    nums = list(nums)
    top = most_common(nums)
    if len(top) > 1 and (top[0][0] - top[1][0] < 2):
        return median_noavg(nums)
    else:
        return median_modes(nums)


def median_noavg(nums):
    nums = sorted(list(nums))
    mid = int(len(nums) / 2)
    return nums[mid]


def median_modes(nums):
    max_count = 0
    counts = {}
    for n in nums:
        if n not in counts:
            counts[n] = 1
        else:
            counts[n] += 1
        max_count = max(max_count, counts[n])
    return median_noavg(k for k, n in counts.items() if max_count - n < 3)


def mean(nums):
    nums = list(nums)
    return sum(nums) / len(nums)


def mean_or_none(nums):
    nums = list(nums)
    if len(nums) == 0:
        return None
    else:
        return sum(nums) / len(nums)


def mean_or_none_round(nums):
    r = mean_or_none(nums)
    if r is None:
        return r
    else:
        return round(r)


def trim(nums, pct=25):
    nums = sorted(list(nums))
    trim_n = int(len(nums) / float(100.0) * pct)
    if trim_n > 0:
        return nums[trim_n:-trim_n]
    else:
        return nums


def most_common(nums):
    counts = {}
    for n in nums:
        if n not in counts:
            counts[n] = 1
        else:
            counts[n] += 1
    return sorted(((counts[n], n) for n in counts), reverse=True)


def most_common_top(nums):
    result = most_common(nums)
    return sorted(item for count, item in result if count == result[0][0])[0]


def error(msg):
    sys.stderr.write("Sniffles2 Error: " + msg + "\n")
    sys.stderr.flush()


def fatal_error(msg):
    error(msg + " (Fatal error, exiting.)")
    exit(1)


def fatal_error_main(msg):
    error(msg + " (Fatal error, exiting.)")
    raise Sniffles2Exit


def load_tandem_repeats(filename, padding):
    contigs_tr = {}
    unsorted = False
    with open(filename, "r") as handle:
        for line in handle:
            parts = line.split("\t")
            if len(parts) >= 3:
                contig, start, end = parts[:3]
                start = int(start)
                end = int(end)
                if contig not in contigs_tr:
                    contigs_tr[contig] = []
                if len(contigs_tr[contig]) > 0:
                    last_start, last_end = contigs_tr[contig][-1]
                    if start < last_start:
                        unsorted = True
                contigs_tr[contig].append((max(0, int(start) - padding), int(end) + padding))

    if unsorted:
        print("Info: The tandem repeat annotations file was not sorted. Sorting it in-memory after loading... (please sort the .bed file once before to save time when running multiple samples)")
        sort_start = time.time()
        for contig in contigs_tr:
            contigs_tr[contig].sort()
        print(f"Info: Optional sorting of input tandem repeat annotations took {time.time() - sort_start:.2f}s.")

    return contigs_tr


center = median_modes



#!/usr/bin/env python3
"""
Population SNF
"""
import logging
import math
from dataclasses import dataclass, asdict
from typing import Optional, Callable

from pysam import VariantFile
try:
    from edlib import align
except ImportError:
    align: Optional[Callable] = None

from sniffles.config import SnifflesConfig
# from sniffles.genotyping import Genotype
from sniffles.snf import SNFileBase
from sniffles.sv import SVCall


log = logging.getLogger(__name__)


@dataclass
class PopulationVariant:
    contig: str
    pos: int
    id: str
    alt: str

    svtype: str
    svlen: int
    end: int

    af: float  # population allele fraction
    genotyped_sample_count: int  # number of genotyped samples for this variant
    variant_sample_count: int  # number of samples in the population carrying this variant

    @staticmethod
    def _calculate_frequency(genotypes: dict[int, tuple]) -> tuple[float, int, int]:
        """
        Calculate population AF from a dict of genotypes. Keys are sample ids, values are genotype tuples.
        Returns a tuple with population AF, total number of genotyped samples and number of samples carrying this SV.
        """
        total_alleles = 0
        variant_alleles = 0
        genotyped_samples = 0
        variant_samples = 0
        gp = SnifflesConfig.GLOBAL.genotype_ploidy

        for gt in genotypes.values():
            if gt[0] == '.':
                continue
            else:
                genotyped_samples += 1
                variant_number = gt[0] + gt[1]
                total_alleles += gp
                variant_alleles += variant_number
                if variant_number > 0:
                    variant_samples += 1

        return variant_alleles / total_alleles, genotyped_samples, variant_samples

    @classmethod
    def from_svcall(cls, svcall: SVCall) -> Optional['PopulationVariant']:
        """
        Generate a population variant from a merged SV call, calculating frequencies.
        """
        af, genotyped_samples, variant_samples = cls._calculate_frequency(svcall.genotypes)
        population_size = len(SnifflesConfig.GLOBAL.snf_input_info)

        if (pct_genotyped := (genotyped_samples / population_size)) < SnifflesConfig.GLOBAL.dev_population_min_gt:
            log.warning(f'Not emitting call {svcall} due to only {genotyped_samples}/{population_size} ({pct_genotyped*100:.2f}%) genotyped samples.')
            return None

        obj = cls(
            contig=svcall.contig,
            pos=svcall.pos,
            id=svcall.id,
            alt=svcall.alt,
            svtype=svcall.svtype,
            svlen=svcall.svlen,
            end=svcall.end,
            af=af,
            genotyped_sample_count=genotyped_samples,
            variant_sample_count=variant_samples,
        )
        return obj

    def match(self, svcall: SVCall) -> int | None:
        """
        Determine if svcall is the same variant. Returns the distance (smaller being better) or None if it
        doesn't match at all.
        """
        config = SnifflesConfig.GLOBAL
        dist = abs(self.pos - svcall.pos) + abs(abs(self.svlen) - abs(svcall.svlen))
        minlen = float(min(abs(self.svlen), abs(svcall.svlen)))
        if dist > config.combine_match * math.sqrt(minlen) or dist > config.combine_match_max:
            return None

        if self.svtype == 'INS' and (limit := config.combine_pctseq):
            distance = align(self.alt, svcall.alt)['editDistance']
            if (self.svlen - distance) / self.svlen <= limit:
                return None

        return dist


@dataclass
class PopulationInfo:
    version: int
    name: str
    description: str
    size: int


class PopulationSNF(SNFileBase):
    """
    A population SNF holds information on variants and their frquency in a population
    """
    _blocks = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._blocks = {}

    def _calculate_block_index(self, pos: int) -> int:
        return int(pos / self.config.snf_block_size) * self.config.snf_block_size

    def get_population_AF(self, svcall: SVCall) -> tuple[float, int] | None:
        """
        Returns the population AF and population size for the given SVCall. Returns None if the variant is not found in the population SNF.
        """
        if svcall.contig not in self._blocks:
            self._blocks[svcall.contig] = self.get_all_blocks(svcall.contig)

        block = str(self._calculate_block_index(svcall.pos))
        best_dist = None
        best_variant = None
        try:
            for pv in self._blocks[svcall.contig][block][svcall.svtype]:
                pv: PopulationVariant
                dist = pv.match(svcall)
                if dist is not None:
                    if best_dist is None or dist < best_dist:
                        best_dist = dist
                        best_variant = pv
        except KeyError:
            ...
        else:
            if best_variant is not None:
                return round(best_variant.af, 5), best_variant.genotyped_sample_count

        return None

    def _create_header(self, config: SnifflesConfig, main_index: dict, snf_candidate_count: int) -> dict:
        """
        Extend the standard snf header with population information
        """
        d = super()._create_header(config, main_index, snf_candidate_count)
        d['population'] = asdict(PopulationInfo(
            version=1,
            name='Population',
            description='A sample population',
            size=len(config.snf_input_info)
        ))
        return d

    def read_header(self):
        """
        Deserialize population information
        """
        super().read_header()
        try:
            self.header['population'] = PopulationInfo(**self.header['population'])
        except:  # noqa
            log.warning(f'Unable to deserialize population information from SNF header.', exc_info=True)

    def _calculate_contig_coverages(self, *args, **kwargs) -> dict:
        """
        Coverages are not used in population SNFs.
        """
        return {}

    def store(self, svcand: SVCall) -> None:
        """
        Convert SVCall to PopulationVariant and store it
        """
        if (variant := PopulationVariant.from_svcall(svcand)) is not None:
            super().store(variant)
        return variant is not None


class PopulationVCF:
    def __init__(self, vcf_path: str):
        self.vcf = VariantFile(vcf_path)

    def get_population_annotation(self, sv: SVCall):
        self.vcf.fetch(sv.contig, sv.pos - 500, sv.pos + sv.svlen + 500)



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     18.10.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#

import os
import sys
import datetime
import argparse
import tempfile
from collections import defaultdict

from typing import Union, Optional

from sniffles import util
from sniffles.region import Region

VERSION = "Sniffles2"
BUILD = "2.6.0"
SNF_VERSION = "S2_rc4"


class ArgFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass


def tobool(v):
    if v is True or v is False:
        return v
    elif v.strip().lower() == "true" or v.strip() == "1":
        return True
    elif v.strip().lower() == "false" or v.strip() == "0":
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value (True | False) required for argument")


class SnifflesConfig(argparse.Namespace):
    GLOBAL: 'SnifflesConfig'

    header = f"Sniffles2: A fast structural variant (SV) caller for long-read sequencing data\n Version {BUILD}\n Contact: sniffles@romanek.at"
    example = """ Usage example A - Call SVs for a single sample:
       sniffles --input sorted_indexed_alignments.bam --vcf output.vcf

       ... OR, with CRAM input and bgzipped+tabix indexed VCF output:
         sniffles --input sample.cram --vcf output.vcf.gz

       ... OR, producing only a SNF file with SV candidates for later multi-sample calling:
         sniffles --input sample1.bam --snf sample1.snf

       ... OR, simultaneously producing a single-sample VCF and SNF file for later multi-sample calling:
         sniffles --input sample1.bam --vcf sample1.vcf.gz --snf sample1.snf

       ... OR, with additional options to specify tandem repeat annotations (for improved call accuracy), reference (for DEL sequences) and mosaic mode for detecting rare SVs:
         sniffles --input sample1.bam --vcf sample1.vcf.gz --tandem-repeats tandem_repeats.bed --reference genome.fa --mosaic

    Usage example B - Multi-sample calling:
       Step 1. Create .snf for each sample: sniffles --input sample1.bam --snf sample1.snf
       Step 2. Combined calling: sniffles --input sample1.snf sample2.snf ... sampleN.snf --vcf multisample.vcf

       ... OR, using a .tsv file containing a list of .snf files, and custom sample ids in an optional second column (one sample per line):
       Step 2. Combined calling: sniffles --input snf_files_list.tsv --vcf multisample.vcf

    Usage example C - Determine genotypes for a set of known SVs (force calling):
       sniffles --input sample.bam --genotype-vcf input_known_svs.vcf --vcf output_genotypes.vcf
       """
    usage = "sniffles --input SORTED_INPUT.bam [--vcf OUTPUT.vcf] [--snf MERGEABLE_OUTPUT.snf] [--threads 4] [--mosaic]\n\n" + header + "\n\n" + example + "\n\n Use --help for full parameter/usage information\n \n"

    parser: argparse.ArgumentParser

    quiet: bool

    dev_call_region: Union[str, dict]

    @property
    def sort(self):
        """
        Output is sorted unless explicitly asked for otherwise, and only if the output is not bgzipped
        """
        return self.vcf_output_bgz or not self.no_sort

    input: str
    vcf: str
    snf: str
    reference: str
    tandem_repeats: str
    phase: bool
    threads: int
    contig: Optional[str]
    run_id: str
    tmp_dir: str

    @property
    def vcf_output_bgz(self) -> Optional[bool]:
        """
        Should the output vcf file be compressed?
        """
        if self.vcf:
            path, ext = os.path.splitext(self.vcf)
            return ext == ".gz" or ext == ".bgz"

    @staticmethod
    def add_main_args(parser):
        main_args = parser.add_argument_group("Common parameters")
        main_args.add_argument("-i", "--input", metavar="IN", type=str, help="For single-sample calling: A coordinate-sorted and indexed .bam/.cram (BAM/CRAM format) file containing aligned reads. - OR - For multi-sample calling: Multiple .snf files (generated before by running Sniffles2 for individual samples with --snf)", required=True, nargs="+")
        main_args.add_argument("-v", "--vcf", metavar="OUT.vcf", type=str, help="VCF output filename to write the called and refined SVs to. If the given filename ends with .gz, the VCF file will be automatically bgzipped and a .tbi index built for it.", required=False)
        main_args.add_argument("--snf", metavar="OUT.snf", type=str, help="Sniffles2 file (.snf) output filename to store candidates for later multi-sample calling", required=False)
        main_args.add_argument("--reference", metavar="reference.fasta", type=str, help="(Optional) Reference sequence the reads were aligned against. To enable output of deletion SV sequences, this parameter must be set.", default=None)
        main_args.add_argument("--tandem-repeats", metavar="IN.bed", type=str, help="(Optional) Input .bed file containing tandem repeat annotations for the reference genome.", default=None)
        main_args.add_argument("--phase", help="Determine phase for SV calls (requires the input alignments to be phased)", default=False, action="store_true")
        main_args.add_argument("-t", "--threads", metavar="N", type=int, help="Number of parallel threads to use (speed-up for multi-core CPUs)", default=4)
        main_args.add_argument("-c", "--contig", default=None, type=str, help="(Optional) Only process the specified contigs. May be given more than once.", action="append")
        main_args.add_argument("--regions", metavar="REGIONS.bed", type=str, help="(Optional) Only process the specified regions.", default=None)
        main_args.add_argument("--tmp-dir", type=str, help="(Optional) Directory where temporary files are written, must exist. If it doesn't, default path is used", default="")

    minsupport: Union[str, int]
    minsvlen: int
    minsvlen_screen_ratio: float
    max_unknown_pct: float

    @staticmethod
    def add_filter_args(parser):
        filter_args = parser.add_argument_group("SV Filtering parameters")
        filter_args.add_argument("--minsupport", metavar="auto", type=str, help="Minimum number of supporting reads for a SV to be reported (default: automatically choose based on coverage)", default="auto")
        filter_args.add_argument("--minsupport-auto-mult", metavar="0.1/0.025", type=float, help="Coverage based minimum support multiplier for germline mode (only for auto minsupport) ", default=None)
        filter_args.add_argument("--minsvlen", metavar="N", type=int, help="Minimum SV length (in bp)", default=50)
        filter_args.add_argument("--minsvlen-screen-ratio", metavar="N", type=float, help="Minimum length for SV candidates (as fraction of --minsvlen)", default=0.9)
        filter_args.add_argument("--mapq", metavar="N", type=int, help="Alignments with mapping quality lower than this value will be ignored", default=argparse.SUPPRESS)
        filter_args.add_argument("--no-qc", "--qc-output-all", help="Output all SV candidates, disregarding quality control steps.", default=False, action="store_true")
        filter_args.add_argument("--qc-stdev", help="Apply filtering based on SV start position and length standard deviation", metavar="True", type=tobool, default=True)
        filter_args.add_argument("--qc-stdev-abs-max", help="Maximum standard deviation for SV length and size (in bp)", metavar="N", type=int, default=500)
        filter_args.add_argument("--qc-strand", help="Apply filtering based on strand support of SV calls", metavar="False", type=tobool, default=False)
        filter_args.add_argument("--qc-coverage", help="Minimum surrounding region coverage of SV calls", metavar="N", type=int, default=1)
        filter_args.add_argument("--long-ins-length", help="Insertion SVs longer than this value are considered as hard to detect based on the aligner and read length and subjected to more sensitive filtering.", metavar="2500", type=int, default=2500)
        filter_args.add_argument("--long-del-length", help="Deletion SVs longer than this value are subjected to central coverage drop-based filtering (Not applicable for --mosaic)", metavar="50000", type=int, default=50000)
        filter_args.add_argument("--long-inv-length", help="Inversion SVs longer than this value are not subjected to central coverage drop-based filtering", metavar="10000", type=int, default=10000)
        filter_args.add_argument("--long-del-coverage", help="Long deletions with central coverage (in relation to upstream/downstream coverage) higher than this value will be filtered (Not applicable for --mosaic)", metavar="0.66", type=float, default=0.66)
        filter_args.add_argument("--long-dup-length", help="Duplication SVs longer than this value are subjected to central coverage increase-based filtering (Not applicable for --mosaic)", metavar="50000", type=int, default=50000)
        filter_args.add_argument("--qc-bnd-filter-strand", help="Filter breakends that do not have support for both strands", type=tobool, default=True)
        filter_args.add_argument("--bnd-min-split-length", help="Minimum length of read splits to be considered for breakends", type=int, default=1000)
        filter_args.add_argument("--long-dup-coverage", help="Long duplications with central coverage (in relation to upstream/downstream coverage) lower than this value will be filtered (Not applicable for --mosaic)", metavar="1.33", type=float, default=1.33)
        filter_args.add_argument("--max-splits-kb", metavar="N", type=float, help="Additional number of splits per kilobase read sequence allowed before reads are ignored", default=0.1)
        filter_args.add_argument("--max-splits-base", metavar="N", type=int, help="Base number of splits allowed before reads are ignored (in addition to --max-splits-kb)", default=3)
        filter_args.add_argument("--min-alignment-length", metavar="N", type=int, help="Reads with alignments shorter than this length (in bp) will be ignored", default=argparse.SUPPRESS)
        filter_args.add_argument("--phase-conflict-threshold", metavar="F", type=float, help="Maximum fraction of conflicting reads permitted for SV phase information to be labelled as PASS (only for --phase)", default=0.1)
        filter_args.add_argument("--detect-large-ins", help="Infer insertions that are longer than most reads and therefore are spanned by few alignments only.", metavar="True", type=tobool, default=True)
        filter_args.add_argument("--max-unknown-pct", help="Maximum percentage of N for an SV to be emitted.", metavar="0.5", type=float, default=0.5)
        # filter_args.add_argument("--large-ins-threshold", metavar="N", type=int, help="Minimum clipping at read ends to be considered a potential large insertion (only with --detect-large-ins)", default=5000)

    cluster_binsize: int
    cluster_binsize_combine_mult: int

    @staticmethod
    def add_cluster_args(parser):
        cluster_args = parser.add_argument_group("SV Clustering parameters")
        cluster_args.add_argument("--cluster-binsize", metavar="N", type=int, help="Initial screening bin size in bp", default=100)
        cluster_args.add_argument("--cluster-r", metavar="R", type=float, help="Multiplier for SV start position standard deviation criterion in cluster merging", default=2.5)
        cluster_args.add_argument("--cluster-repeat-h", metavar="H", type=float, help="Multiplier for mean SV length criterion for tandem repeat cluster merging", default=1.5)
        cluster_args.add_argument("--cluster-repeat-h-max", metavar="N", type=float, help="Max. merging distance based on SV length criterion for tandem repeat cluster merging", default=1000)
        cluster_args.add_argument("--cluster-merge-pos", metavar="N", type=int, help="Max. merging distance for insertions and deletions on the same read and cluster in non-repeat regions", default=150)
        cluster_args.add_argument("--cluster-merge-len", metavar="F", type=float, help="Max. size difference for merging SVs as fraction of SV length", default=0.33)
        cluster_args.add_argument("--cluster-merge-bnd", metavar="N", type=int, help="Max. merging distance for breakend SV candidates.", default=1000)

    genotype_ploidy: int
    genotype_vcf: str

    @staticmethod
    def add_genotype_args(parser):
        genotype_args = parser.add_argument_group("SV Genotyping parameters")
        genotype_args.add_argument("--genotype-ploidy", metavar="N", type=int, help="Sample ploidy (currently fixed at value 2)", default=2)
        genotype_args.add_argument("--genotype-error", metavar="N", type=float, help="Estimated false positive rate for leads (relating to total coverage)", default=0.05)
        genotype_args.add_argument("--sample-id", type=str, help="Custom ID for this sample, used for later multi-sample calling (stored in .snf)", default=None)
        genotype_args.add_argument("--genotype-vcf", metavar="IN.vcf", type=str, help="Determine the genotypes for all SVs in the given input .vcf file (forced calling). Re-genotyped .vcf will be written to the output file specified with --vcf.", default=None)

    combine_high_confidence: float
    combine_low_confidence: float
    combine_low_confidence_abs: int
    combine_null_min_coverage: int
    combine_match: int
    combine_match_max: int
    combine_separate_intra: bool
    combine_output_filtered: bool
    combine_pair_relabel: bool
    combine_pair_relabel_threshold: int
    combine_close_handles: bool
    combine_pctseq: float
    combine_max_inmemory_results: int
    combine_support_threshold: int
    combine_population: str

    @classmethod
    def add_multi_args(cls, parser):
        """
        Arguments for multi-sample calling
        """        
        multi_args = parser.add_argument_group("Multi-Sample Calling / Combine parameters")
        multi_args.add_argument("--combine-high-confidence", metavar="F", type=float, help="Minimum fraction of samples in which a SV needs to have individually passed QC for it to be reported in combined output (a value of zero will report all SVs that pass QC in at least one of the input samples)", default=0.0)
        multi_args.add_argument("--combine-low-confidence", metavar="F", type=float, help="Minimum fraction of samples in which a SV needs to be present (failed QC) for it to be reported in combined output", default=0.2)
        multi_args.add_argument("--combine-low-confidence-abs", metavar="N", type=int, help="Minimum absolute number of samples in which a SV needs to be present (failed QC) for it to be reported in combined output", default=2)
        multi_args.add_argument("--combine-null-min-coverage", metavar="N", type=int, help="Minimum coverage for a sample genotype to be reported as 0/0 (sample genotypes with coverage below this threshold at the SV location will be output as ./.)", default=5)
        multi_args.add_argument("--combine-match", metavar="N", type=int, help="Multiplier for maximum deviation of multiple SV's start/end position for them to be combined across samples. Given by max_dev=M*sqrt(min(SV_length_a,SV_length_b)), where M is this parameter.", default=250)
        multi_args.add_argument("--combine-match-max", metavar="N", type=int, help="Upper limit for the maximum deviation computed for --combine-match, in bp.", default=1000)
        multi_args.add_argument("--combine-separate-intra", help="Disable combination of SVs within the same sample", default=False, action="store_true")
        multi_args.add_argument("--combine-output-filtered", help="Include low-confidence / mosaic SVs in multi-calling", default=False, action="store_true")
        multi_args.add_argument("--combine-pair-relabel", help="Override low-quality genotypes when combining 2 samples (may be used for e.g. tumor-normal comparisons)", default=False, action="store_true")
        multi_args.add_argument("--combine-pair-relabel-threshold", help="Genotype quality below which a genotype call will be relabeled", default=20, type=int)
        multi_args.add_argument("--combine-close-handles", help="Close .SNF file handles after each use. May lower performance, but may be required when maximum number of file handles supported by OS is reached when merging many samples.", default=False, action="store_true")
        multi_args.add_argument("--combine-pctseq", default=0.7, type=float, help="Minimum alignment distance as percent of SV length to be merged. Set to 0 to disable alignments for merging.")
        multi_args.add_argument("--combine-max-inmemory-results", default=20, type=int, help=argparse.SUPPRESS)
        multi_args.add_argument("--combine-support-threshold", default=3, metavar="N", type=int, help="Minimum support for SVs to be considered for multi-sample calling.")
        multi_args.add_argument("--combine-population", metavar="population.snf", type=str, help="Name of a population SNF to enable population annotation.")
        multi_args.add_argument("--re-qc", metavar="auto", default="auto", type=str, help="Re-QC SVs from SNF files. Set to 0 to disable re-qc of SNF files. Set to 1 to force re-qc. Default of 'auto' will try to fix known errors in SNF files.")

        # multi_args.add_argument("--combine-exhaustive", help="(DEV) Disable performance optimization in multi-calling", default=False, action="store_true")
        # multi_args.add_argument("--combine-relabel-rare", help="(DEV)", default=False, action="store_true")
        # multi_args.add_argument("--combine-with-missing", help="(DEV)", default=False, action="store_true")

    allow_overwrite: bool

    @staticmethod
    def add_postprocess_args(parser):
        """
        Postprocessing arguments
        """
        postprocess_args = parser.add_argument_group("SV Postprocessing, QC and output parameters")
        postprocess_args.add_argument("--output-rnames", help="Output names of all supporting reads for each SV in the RNAMEs info field", default=False, action="store_true")
        postprocess_args.add_argument("--no-consensus", help="Disable consensus sequence generation for insertion SV calls (may improve performance)", default=False, action="store_true")
        postprocess_args.add_argument("--no-sort", help="Do not sort output VCF by genomic coordinates (may slightly improve performance)", default=False, action="store_true")
        postprocess_args.add_argument("--no-progress", help="Disable progress display", default=False, action="store_true")
        postprocess_args.add_argument("--quiet", help="Disable all logging, except errors", default=False, action="store_true")
        postprocess_args.add_argument("--max-del-seq-len", metavar="N", type=int, help="Maximum deletion sequence length to be output. Deletion SVs longer than this value will be written to the output as symbolic SVs.", default=50000)
        postprocess_args.add_argument("--symbolic", help="Output all SVs as symbolic, including insertions and deletions, instead of reporting nucleotide sequences.", default=False, action="store_true")
        postprocess_args.add_argument("--allow-overwrite", help="Allow overwriting output files if already existing", default=False, action="store_true")

    mosaic_include_germline: bool
    mosaic_qc_nm: bool
    # TODO some better rules here
    mosaic_min_reads: int = 3
    mosaic_use_strand_thresholds: int = 10

    @staticmethod
    def add_mosaic_args(parser):
        mosaic_args = parser.add_argument_group("Mosaic calling mode parameters")
        mosaic_args.add_argument("--mosaic", help="Set Sniffles run mode to detect rare, somatic and mosaic SVs", default=False, action="store_true")
        mosaic_args.add_argument("--mosaic-af-max", help="Maximum allele frequency for which SVs are considered mosaic", metavar="F", default=0.218, type=float)
        mosaic_args.add_argument("--mosaic-af-min", help="Minimum allele frequency for mosaic SVs to be output", metavar="F", default=0.05, type=float)
        mosaic_args.add_argument("--mosaic-qc-invdup-min-length", help="Minimum SV length for mosaic inversion and duplication SVs", metavar="N", default=500, type=int)
        mosaic_args.add_argument("--mosaic-qc-nm", default=True, action="store_true", help=argparse.SUPPRESS)
        mosaic_args.add_argument("--mosaic-qc-nm-mult", metavar="F", type=float, default=1.66, help=argparse.SUPPRESS)
        mosaic_args.add_argument("--mosaic-qc-coverage-max-change-frac", help="Maximum relative coverage change across SV breakpoints", metavar="F", type=float, default=0.1)
        mosaic_args.add_argument("--mosaic-qc-strand", help="Apply filtering based on strand support of SV calls", metavar="True", type=tobool, default=True)
        mosaic_args.add_argument("--mosaic-include-germline", help="Report germline SVs as well in mosaic mode", default=False, action="store_true")

    qc_nm: bool
    combine_consensus: bool
    low_memory: bool
    dev_population_snf: str
    dev_population_min_gt: float
    dev_debug: int

    @staticmethod
    def add_developer_args(parser):
        """
        Developer arguments
        """
        developer_args = parser.add_argument_group("Developer parameters")

        developer_args.add_argument("--dev-emit-sv-lengths", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-cache", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-cache-dir", metavar="PATH", type=str, default=None, help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-debug-svtyping", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-keep-lowqual-splits", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-call-region", metavar="REGION", type=str, default=None, help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-dump-clusters", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-merge-inline", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-seq-cache-maxlen", metavar="N", type=int, default=50000, help=argparse.SUPPRESS)
        developer_args.add_argument("--consensus-max-reads", metavar="N", type=int, default=20, help=argparse.SUPPRESS)
        developer_args.add_argument("--consensus-max-reads-bin", metavar="N", type=int, default=10, help=argparse.SUPPRESS)
        developer_args.add_argument("--combine-consensus", help="Output the consensus genotype of all samples", default=False, action="store_true")
        developer_args.add_argument("--dev-dump-coverage", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-no-resplit", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-no-resplit-repeat", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-skip-snf-validation", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--low-memory", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--repeat", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--qc-nm", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--qc-nm-mult", metavar="F", type=float, default=1.66, help=argparse.SUPPRESS)
        developer_args.add_argument("--qc-coverage-max-change-frac", help="Maximum relative coverage change across SV breakpoints", metavar="F", type=float, default=-1)
        developer_args.add_argument("--coverage-updown-bins", metavar="N", type=int, default=5, help=argparse.SUPPRESS)
        developer_args.add_argument("--coverage-shift-bins", metavar="N", type=int, default=3, help=argparse.SUPPRESS)
        developer_args.add_argument("--coverage-shift-bins-min-aln-length", metavar="N", type=int, default=1000, help=argparse.SUPPRESS)
        developer_args.add_argument("--cluster-binsize-combine-mult", metavar="N", type=int, default=5, help=argparse.SUPPRESS)
        developer_args.add_argument("--cluster-resplit-binsize", metavar="N", type=int, default=20, help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-trace-read", default=False, metavar="read_id", type=str, help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-split-max-query-distance-mult", metavar="N", type=int, default=5, help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-no-qc", default=False, action="store_true", help=argparse.SUPPRESS)  # noqc + mapq0 + minAlnLen0
        developer_args.add_argument("--dev-disable-interblock-threads", default=False, help=argparse.SUPPRESS, action="store_true")
        developer_args.add_argument("--dev-combine-medians", default=False, help=argparse.SUPPRESS, action="store_true")
        developer_args.add_argument("--dev-monitor-memory", metavar="N", type=int, default=0, help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-monitor-filename", metavar="memory.csv", type=str, help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-debug-log", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-progress-log", default=False, action="store_true", help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-population-snf", metavar="population.snf", type=str, help=argparse.SUPPRESS)
        developer_args.add_argument("--dev-population-min-gt", default=0.75, type=float, help=argparse.SUPPRESS)  # min
        developer_args.add_argument("--dev-debug", default=0, type=int, help=argparse.SUPPRESS)  # Enable debug connection on the given port

        # developer_args.add_argument("--qc-strand", help="(DEV)", default=False, action="store_true")

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)

        parser = argparse.ArgumentParser(description="", epilog=self.example, formatter_class=lambda prog: ArgFormatter(prog, max_help_position=100, width=150), usage=self.usage)
        parser.add_argument("--version", action="version", version=f"Sniffles2, Version {BUILD}")

        self.add_main_args(parser)
        self.add_filter_args(parser)
        self.add_cluster_args(parser)
        self.add_genotype_args(parser)
        self.add_multi_args(parser)
        self.add_postprocess_args(parser)
        self.add_mosaic_args(parser)
        self.add_developer_args(parser)

        parser.parse_args(args=args or None, namespace=self)

        if not self.tmp_dir or not os.path.exists(self.tmp_dir):
            self.tmp_dir = tempfile.gettempdir()

        if self.quiet:
            sys.stdout = open(os.devnull, "w")

        self.start_date = datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S")
        self.run_id = f'{os.environ.get("SLURM_JOB_ID") or os.getpid()}'

        self.task_count_multiplier = 0

        self.version = VERSION
        self.build = BUILD
        self.snf_format_version = SNF_VERSION
        self.command = " ".join(sys.argv)

        if self.dev_call_region is not None:
            region_contig, region_startend = self.dev_call_region.replace(",", "").split(":")
            start, end = region_startend.split("-")
            self.dev_call_region = dict(contig=region_contig, start=int(start), end=int(end))

        if self.contig and self.regions:
            util.fatal_error('Please provide either --contig or --regions, not both.')

        if self.regions is not None:
            regions = defaultdict(list)
            with open(self.regions, 'r') as f:
                for line in f.readlines():
                    if line.startswith('#') or line.strip() == '':
                        continue
                    else:
                        r = Region.from_bed_line(line)
                        if r is not None:
                            regions[r.contig].append(r)
            self.regions_by_contig = regions
        else:
            self.regions_by_contig = {}

        # "--minsvlen" parameter is for final output filtering
        # for intermediate steps, a lower threshold is used to account for sequencing, mapping imprecision
        self.minsvlen_screen = int(self.minsvlen_screen_ratio * self.minsvlen)
        # config.minsupport_screen=max(1,int(0.333*config.minsupport*(config.cluster_binsize/100.0)))

        if self.minsupport != "auto":
            self.minsupport = int(self.minsupport)

        if self.dev_no_qc:
            self.no_qc = True

        if self.re_qc == 'auto':
            self.reqc = 'auto'
        elif self.re_qc in ('0', '1'):
            self.reqc = bool(int(self.re_qc))
        else:
            util.fatal_error('Invalid value for --re-qc, allowed values are: auto, 0, 1')

        if not hasattr(self, 'mapq'):
            self.mapq = 0 if self.dev_no_qc else 20
        if not hasattr(self, 'min_alignment_length'):
            self.min_alignment_length = 0 if self.dev_no_qc else 1000

        # --minsupport auto defaults
        self.minsupport_auto_base = 1.5
        self.minsupport_auto_regional_coverage_weight = 0.75

        if self.minsupport_auto_mult is None:
            self.minsupport_auto_mult = 0.1

        self.coverage_binsize = self.cluster_binsize
        self.coverage_binsize_combine = self.cluster_binsize * self.cluster_binsize_combine_mult

        # INS Consensus parameters
        # config.consensus_max_reads=20
        # config.consensus_max_reads_bin=10
        self.consensus_min_reads = 4
        self.consensus_kmer_len = 6
        self.consensus_kmer_skip_base = 3
        self.consensus_kmer_skip_seqlen_mult = 1.0 / 500.0
        self.consensus_low_threshold = 0.0  # 0.15

        # Large INS
        self.long_ins_rescale_base = 1.66
        self.long_ins_rescale_mult = 0.33

        # BND
        self.bnd_cluster_length = 1000

        # Genotyping
        self.genotype_format = "GT:GQ:DR:DV"
        self.genotype_none = (".", ".", 0, 0, 0, (None, None))
        self.genotype_null = (0, 0, 0, 0, 0, (None, None))
        self.genotype_min_z_score = 5
        if self.genotype_ploidy != 2:
            util.fatal_error("Currently only --genotype-ploidy 2 is supported")

        # SNF
        self.snf_block_size = 10 ** 5

        # Combine
        self.combine_exhaustive = False
        self.combine_relabel_rare = False
        self.combine_overlap_abs = 2500
        self.combine_min_size = 100

        # Misc
        self.precise = 25  # Max. sum of pos and length stdev for SVs to be labelled PRECISE
        self.tandem_repeat_region_pad = 500
        self.id_prefix = "Sniffles2."
        self.phase_identifiers = ["1", "2"]

        self.dev_profile = False

        self.workdir = os.getcwd()

        # Mosaic
        if self.mosaic_include_germline:
            self.mosaic = True

        self.qc_nm_measure = self.qc_nm
        if self.mosaic:
            # config.qc_coverage_max_change_frac=config.mosaic_qc_coverage_max_change_frac
            self.qc_nm_measure = self.qc_nm_measure or self.mosaic_qc_nm
            # config.qc_nm_mult=config.mosaic_qc_nm_mult
            # config.qc_strand=config.mosaic_qc_strand

        SnifflesConfig.GLOBAL = self



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     15.08.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
import logging
from dataclasses import dataclass
from typing import Optional, Callable

try:
    from edlib import align
except ImportError:
    logging.getLogger('sniffles.dependencies').info('Dependency edlib not installed - alignments are disabled.')
    align: Optional[Callable] = None

from sniffles import util

TYPES = ["INS", "DEL", "DUP", "INV", "BND"]


@dataclass
class SVCallBNDInfo:
    mate_contig: str
    mate_ref_start: int
    is_first: bool
    is_reverse: bool


@dataclass
class SVCallPostprocessingInfo:
    cluster: list


@dataclass
class SVCall:
    contig: str
    pos: int
    id: str
    ref: str
    alt: str
    qual: int
    filter: str
    info: dict

    svtype: str
    svlen: int
    end: int
    genotypes: dict[int, tuple]

    precise: bool
    support: int
    rnames: list[str]

    qc: bool
    nm: float
    postprocess: Optional[SVCallPostprocessingInfo]

    svlens: list[int] = None

    fwd: int = None
    rev: int = None

    coverage_upstream: int = 0
    coverage_downstream: int = 0
    coverage_start: int = 0
    coverage_center: int = 0
    coverage_end: int = 0

    sample_internal_id: int = None
    bnd_info: SVCallBNDInfo = None

    raw_vcf_line: Optional[str] = None
    raw_vcf_line_index: Optional[int] = None

    _counter = 0

    def set_info(self, k, v):
        self.info[k] = v

    def get_info(self, k):
        return self.info[k] if k in self.info else None

    def has_info(self, k):
        return k in self.info

    def finalize(self):
        self.postprocess = None


@dataclass
class SVGroup:
    """
    For combining grouped SV calls from multiple .snf samples into one SV call
    """
    candidates: list[SVCall]
    pos_mean: float
    len_mean: float
    included_samples: set
    coverages_nonincluded: dict

    bnd_mate_ref_start_mean: float = None

    # _pos_mean: float = None
    # _len_mean: float = None

    _counter = 0

    def __new__(cls, *args, **kwargs):
        cls._counter += 1
        return super().__new__(cls)

    def __del__(self):
        SVGroup._counter -= 1

    # @property
    # def pos_mean(self) -> float:
    #     if self._pos_mean is None:
    #         self._pos_mean = util.mean(c.pos for c in self.candidates)
    #     return self._pos_mean
    #
    # @property
    # def len_mean(self) -> float:
    #     if self._len_mean is None:
    #         self._len_mean = util.mean(c.svlen for c in self.candidates)
    #     return self._len_mean

    @classmethod
    def from_candidate(cls, candidate: SVCall) -> "SVGroup":
        """
        Start a new group from given candidate.
        """
        obj = cls(
            candidates=[candidate],
            pos_mean=float(candidate.pos),
            len_mean=float(abs(candidate.svlen)),
            included_samples={candidate.sample_internal_id},
            coverages_nonincluded=dict()
        )
        if candidate.svtype == "BND":
            obj.bnd_mate_contig = candidate.bnd_info.mate_contig
            obj.bnd_mate_ref_start_mean = candidate.bnd_info.mate_ref_start
        return obj

    def align_call(self, candidate: SVCall, limit: float) -> bool:
        """
        Check if the candidate aligns with this group better than given limit.
        """
        if not limit or align is None:
            return True

        distance = align(self.candidates[0].alt, candidate.alt)['editDistance']

        return ((self.len_mean-distance) / self.len_mean) > limit

    def check_call(self):
        """
        Checks if this group is a fit for the given candidate, and adds the candidate if it is. Returns true
        if the candidate was added to this group, false otherwise.
        """

    def add_candidate(self, candidate: SVCall):
        """
        Adds a candidate to this group, updating mean position, length
        and optionally bnd ref start.
        """
        group_size = len(self.candidates)
        self.pos_mean *= group_size
        self.len_mean *= group_size
        self.pos_mean += candidate.pos
        self.len_mean += abs(candidate.svlen)
        if candidate.svtype == "BND":
            self.bnd_mate_ref_start_mean *= group_size
            self.bnd_mate_ref_start_mean += candidate.bnd_info.mate_ref_start

        self.candidates.append(candidate)
        group_size += 1
        self.pos_mean /= group_size
        self.len_mean /= group_size
        self.included_samples.add(candidate.sample_internal_id)

        if candidate.svtype == "BND":
            self.bnd_mate_ref_start_mean /= group_size

    def call(self, config, task) -> Optional[SVCall]:
        """
        Call this group, returning either an SVCall or None.
        """
        first_cand = self.candidates[0]

        # Filtering
        samples_count = float(len(config.snf_input_info))
        n_samples = len(config.snf_input_info)
        sample_internal_ids = set(sample["internal_id"] for sample in config.snf_input_info)
        total_count = len(self.included_samples)
        pass_count = sum(cand.qc for cand in self.candidates)
        qc = ((pass_count > 0 and pass_count / samples_count >= config.combine_high_confidence) or
              (total_count / samples_count >= config.combine_low_confidence and
               total_count >= config.combine_low_confidence_abs))

        if not qc:
            if config.no_qc and n_samples == 1:
                pass
            else:
                return None

        if not config.combine_output_filtered and not any(cand.qc and cand.filter == "PASS" for cand in self.candidates):
            if config.no_qc and n_samples == 1:
                pass
            else:
                return None

        rnames = [] if config.output_rnames else None
        genotypes = {}

        for cand in self.candidates:
            if rnames is not None and cand.rnames is not None:
                rnames.extend(cand.rnames)

            if 0 not in cand.genotypes:
                cand.genotypes[0] = (".", ".", 0, 0, cand.support, (None, None))
            if cand.sample_internal_id in genotypes:
                # Intra-sample merging
                a, b, gt_qual, dr, dv, ps = cand.genotypes[0]
                curr_a, curr_b, curr_gt_qual, curr_dr, curr_dv, curr_ps, curr_id = genotypes[cand.sample_internal_id]
                new_id = curr_id + "," + config.id_prefix + cand.id
                if (curr_a == ".") or (a != "." and (a, b) >= (curr_a, curr_b)):
                    genotypes[cand.sample_internal_id] = (a, b, gt_qual, dr, dv, ps, new_id)
                else:
                    genotypes[cand.sample_internal_id] = (curr_a, curr_b, curr_gt_qual, curr_dr, curr_dv, curr_ps, new_id)
            else:
                a, b, gt_qual, dr, dv, ps = cand.genotypes[0]
                genotypes[cand.sample_internal_id] = (a, b, gt_qual, dr, dv, ps, config.id_prefix + cand.id)

        for sample_internal_id in sample_internal_ids:
            if sample_internal_id in genotypes:
                continue
            coverage = self.coverages_nonincluded[sample_internal_id]
            if coverage >= config.combine_null_min_coverage:
                genotypes[sample_internal_id] = (0, 0, 0, coverage, 0, (None, None), "NULL")
            else:
                genotypes[sample_internal_id] = (".", ".", 0, coverage, 0, (None, None), "NULL")

        if config.combine_consensus:
            genotypes_consensus = {}
            for a, b, gt_qual, dr, dv in genotypes.values():
                if not (a, b) in genotypes_consensus:
                    genotypes_consensus[(a, b)] = {"count": 0, "qual": list(), "dr": list(), "dv": list()}
                genotypes_consensus[(a, b)]["count"] += 1
                genotypes_consensus[(a, b)]["qual"].append(gt_qual)
                genotypes_consensus[(a, b)]["dr"].append(dr)
                genotypes_consensus[(a, b)]["dv"].append(dv)
            most_common_count = genotypes_consensus[sorted(genotypes_consensus,
                                                           key=lambda k: genotypes_consensus[k]["count"],
                                                           reverse=True)[0]]["count"]
            most_common_gt = [gt for gt in genotypes_consensus if genotypes_consensus[gt]["count"] == most_common_count]
            cons_a, cons_b = max(most_common_gt)
            consensus_info = genotypes_consensus[(cons_a, cons_b)]
            genotypes = {0: (cons_a, cons_b, int(sum(consensus_info["qual"]) / consensus_info["count"]),
                             sum(consensus_info["dr"]), sum(consensus_info["dv"]))}
            if cons_a != 1 and cons_b != 1:
                if config.no_qc and n_samples == 1:
                    pass
                else:
                    return None

        if config.combine_pair_relabel:
            max_gt = (0, 0)
            for sample_id in genotypes:
                a, b, qual, dr, dv, ps, new_id = genotypes[sample_id]
                if qual > config.combine_pair_relabel_threshold and a != ".":
                    max_gt = max(max_gt, (a, b))

            if max_gt != (0, 0):
                for sample_id in genotypes:
                    a, b, qual, dr, dv, ps, new_id = genotypes[sample_id]
                    if qual < config.combine_pair_relabel_threshold and a != ".":
                        max_a, max_b = max_gt
                        genotypes[sample_id] = (max_a, max_b, qual, dr, dv, ps, new_id)

        svcall_pos = int(util.median(cand.pos for cand in self.candidates))
        svcall_svlen = int(util.median(cand.svlen for cand in self.candidates))

        if config.dev_emit_sv_lengths:
            svcall_svlens: list[int] | None = [length in cand_lengths
                                               for cand_lengths in self.candidates
                                               for length in cand_lengths.svlens]
        else:
            svcall_svlens = None

        svcall_alt = first_cand.alt
        svcall_alt_mindist = abs(len(svcall_alt) - svcall_svlen)
        if first_cand.svtype == "INS":
            svcall_end = svcall_pos
            for cand in self.candidates:
                dist = abs(len(cand.alt) - svcall_svlen)
                if dist < svcall_alt_mindist:
                    svcall_alt_mindist = dist
                    svcall_alt = cand.alt
        else:
            svcall_end = svcall_pos + abs(svcall_svlen)

        svcall = SVCall(contig=first_cand.contig,
                        pos=svcall_pos if config.dev_combine_medians else first_cand.pos,
                        id=f"{first_cand.svtype}.{task.sv_id:X}M{task.id:X}",
                        ref="N",
                        alt=svcall_alt,
                        qual=util.mean_or_none_round(int(cand.qual) for cand in self.candidates if cand.qual is not None),
                        filter="PASS" if n_samples != 1 else first_cand.filter,
                        info=dict() if n_samples != 1 else first_cand.info,
                        svtype=first_cand.svtype,
                        svlen=svcall_svlen if config.dev_combine_medians else first_cand.svlen,
                        svlens=svcall_svlens,
                        end=svcall_end if config.dev_combine_medians else first_cand.end,
                        genotypes=genotypes,
                        precise=sum(int(cand.precise) for cand in self.candidates) / float(len(self.candidates)) > 0.5,
                        support=round(util.mean(cand.support for cand in self.candidates)),
                        rnames=rnames,
                        postprocess=None,
                        qc=True,
                        nm=-1,
                        fwd=sum(cand.fwd for cand in self.candidates),
                        rev=sum(cand.rev for cand in self.candidates),
                        coverage_upstream=util.mean_or_none_round(cand.coverage_upstream for cand in self.candidates if cand.coverage_upstream is not None),
                        coverage_start=util.mean_or_none_round(cand.coverage_start for cand in self.candidates if cand.coverage_start is not None),
                        coverage_center=util.mean_or_none_round(cand.coverage_center for cand in self.candidates if cand.coverage_center is not None),
                        coverage_end=util.mean_or_none_round(cand.coverage_end for cand in self.candidates if cand.coverage_end is not None),
                        coverage_downstream=util.mean_or_none_round(cand.coverage_downstream for cand in self.candidates if cand.coverage_downstream is not None))

        if n_samples != 1:
            svcall.set_info("STDEV_POS", util.stdev(cand.pos for cand in self.candidates))
            svcall.set_info("STDEV_LEN", util.stdev(cand.svlen for cand in self.candidates))

        if abs(svcall.svlen) < config.minsvlen_screen:
            return None

        task.sv_id += 1

        if psnf := config.combine_population:
            pop_af = psnf.get_population_AF(svcall)
            af, sz = pop_af if pop_af is not None else (None, None)
            svcall.set_info("POPULATION_VAF", af)
            svcall.set_info("POPULATION_SIZE", sz)

        return svcall


def calculate_bounds(svtype, ref_start_mode, svlen_mode):
    if svtype == "INS":
        svstart = ref_start_mode
        svend = ref_start_mode
    elif svtype == "DEL":
        svstart = ref_start_mode + svlen_mode
        svend = ref_start_mode
    else:
        svstart = ref_start_mode
        svend = svstart + abs(svlen_mode)
    return svstart, svend


def call_from(cluster, config, keep_qc_fails, task):
    leads = cluster.leads

    svtype = cluster.svtype
    stdev_pos, stdev_len = None, None
    qc = True

    svlen = util.center(v.svlen for v in leads)

    if config.dev_emit_sv_lengths:
        svlens = [v.svlen for v in leads]
    else:
        svlens = None

    if abs(svlen) < config.minsvlen_screen:
        return

    # Count inline events only once per read, but split events as individual alignments, as in coverage calculation
    # inline_qnames=set(k.read_qname for k in leads if k.source=="INLINE")
    # support=len(inline_qnames)+sum(1 for k in leads if k.source!="INLINE")
    support_set = set(k.read_qname for k in leads)
    if svtype == "INS" and svlen >= config.long_ins_length:
        support_long_set = set(lead.read_qname for lead in cluster.leads_long)
        support_long = len(support_long_set)
        support_set |= support_long_set
        support = len(support_set)
    else:
        support = len(support_set)
        support_long = 0
    ref_start = util.center(v.ref_start for v in leads)
    stdev_pos = util.stdev(util.trim((v.ref_start for v in leads)))
    if svtype != "BND":
        stdev_len = util.stdev(util.trim((v.svlen for v in leads)))
        precise = (stdev_pos + stdev_len < config.precise)
    else:
        precise = (stdev_pos < config.precise)
    svstart, svend = calculate_bounds(svtype, ref_start, svlen)
    qual = int(util.mean(v.mapq for v in leads))

    support_fwd = sum(lead.strand == "+" for lead in leads)
    support_rev = len(leads) - support_fwd

    filter_sv = "PASS"

    if config.qc_nm_measure:
        nm_mean = util.mean(v.nm for v in leads)
    else:
        nm_mean = -1

    if not keep_qc_fails and not qc:
        return

    svpi = SVCallPostprocessingInfo(cluster=cluster)

    if config.output_rnames or config.snf is not None:
        rnames = list(support_set)
    else:
        rnames = None

    svcall = SVCall(contig=cluster.contig,
                    pos=svstart,
                    id=f"{svtype}.{task.sv_id:X}S{task.id:X}",
                    ref="N",
                    alt=f"<{svtype}>",
                    qual=qual,
                    filter=filter_sv,
                    info=dict(),
                    svtype=svtype,
                    svlen=svlen,
                    svlens=svlens,
                    end=svend,
                    genotypes=dict(),
                    precise=precise,
                    support=support,
                    rnames=rnames,
                    postprocess=svpi,
                    qc=qc,
                    nm=nm_mean,
                    fwd=support_fwd,
                    rev=support_rev)

    if svtype == "BND":
        resolve_bnd(svcall, cluster, config)
    elif svtype == "INS":
        svcall.set_info("SUPPORT_LONG", support_long)
    elif svtype == "DEL":
        svcall.set_info("SUPPORT_SA", len([l for l in leads if l.source != "INLINE"]))

    if stdev_pos is not None:
        svcall.set_info("STDEV_POS", stdev_pos)
    if stdev_len is not None:
        svcall.set_info("STDEV_LEN", stdev_len)

    # svcall.set_info("CLUSTER_ID",cluster.id)

    task.sv_id += 1
    yield svcall


def merge_inner_bounds(leads, config):
    svlength_binsize = 25
    read_svlengths = {}
    read_starts = {}
    for ld in leads:
        if not ld.read_id in read_svlengths:
            read_svlengths[ld.read_id] = 0
            read_starts[ld.read_id] = ld.ref_start
        read_svlengths[ld.read_id] += ld.svlen
        read_starts[ld.read_id] = min(ld.ref_start, read_starts[ld.read_id])
    svlen = util.center(v for k, v in read_svlengths.items())
    pos = util.center(v for k, v in read_starts.items())

    svlengths_reads = {}
    for read_id, length in read_svlengths.items():
        bin = int(length / svlength_binsize) * svlength_binsize
        if not bin in svlengths_reads:
            svlengths_reads[bin] = [(read_id, length)]
        else:
            svlengths_reads[bin].append((read_id, length))

    return pos, svlen, util.stdev(util.trim((v for k, v in read_starts.items()))), util.stdev(util.trim((v for k, v in read_svlengths.items())))


def resolve_bnd(svcall, cluster, config):
    mate_contig = util.most_common_top([lead.bnd_info.mate_contig for lead in cluster.leads])
    selected = [lead for lead in cluster.leads if lead.bnd_info.mate_contig == mate_contig]
    mate_ref_start = util.center([lead.bnd_info.mate_ref_start for lead in selected])
    is_first = util.most_common_top([lead.bnd_info.is_first for lead in selected])
    is_reverse = util.most_common_top([lead.bnd_info.is_reverse for lead in selected])
    svcall.alt = (("N" if is_first else "") +
                  ("]" if is_reverse else "[") +
                  f"{mate_contig}:{mate_ref_start}" +
                  ("]" if is_reverse else "[") +
                  ("N" if not is_first else ""))
    svcall.support = len(set(k.read_qname for k in selected))
    cluster.leads = selected
    svcall.bnd_info = SVCallBNDInfo(mate_contig=mate_contig, mate_ref_start=mate_ref_start, is_first=is_first, is_reverse=is_reverse)
    svcall.set_info("CHR2", mate_contig)


def call_groups(svgroups: list[SVGroup], config, task):
    for group in svgroups:
        svcall = group.call(config, task)
        if svcall is not None:
            yield svcall


def classify_splits(read, leads, config, main_contig) -> list:
    """
    Determines the SV type of a split read (read with supplementary alignments). Returns (possibly changed) list of leads.
    """
    minsvlen_screen = config.minsvlen_screen
    min_split_len_bnd = config.bnd_min_split_length

    leads.sort(key=lambda ld: ld.qry_start)
    last = leads[0]
    last.svtypes_starts_lens = []
    hints = 0

    if last.qry_start >= config.long_ins_length * 0.5:
        last.svtypes_starts_lens.append(("INS", last.ref_start, None))

    for i in range(1, len(leads)):
        curr = leads[i]
        curr.svtypes_starts_lens = []

        if curr.contig == last.contig:
            rev = (curr.strand == "-")
            fwd = not rev
            if curr.strand == last.strand:
                #
                # INS, DEL, DUP
                #
                if (fwd and (curr.qry_start - last.qry_end) >= minsvlen_screen
                        and (curr.qry_start - last.qry_end) - (curr.ref_start - last.ref_end) >= minsvlen_screen):
                    # INS, FWD
                    svstart = curr.ref_start
                    svlen = (curr.qry_start - last.qry_end)
                    if svlen <= config.dev_seq_cache_maxlen:
                        curr.seq = read.query_sequence[last.qry_end:curr.qry_start]
                    else:
                        curr.seq = None
                    curr.svtypes_starts_lens.append(("INS", svstart, svlen))
                    hints += 1

                elif (rev and (curr.qry_start - last.qry_end) >= minsvlen_screen
                      and (curr.qry_start - last.qry_end) - (last.ref_start - curr.ref_end) >= minsvlen_screen):
                    # INS, REV
                    svstart = last.ref_start
                    svlen = (curr.qry_start - last.qry_end)
                    if svlen <= config.dev_seq_cache_maxlen:
                        curr.seq = read.query_sequence[last.qry_end:curr.qry_start]
                    else:
                        curr.seq = None
                    curr.svtypes_starts_lens.append(("INS", svstart, svlen))
                    hints += 1

                elif (fwd and (curr.ref_start - last.ref_end) >= minsvlen_screen
                      and (curr.ref_start - last.ref_end) - (curr.qry_start - last.qry_end) >= minsvlen_screen):
                    # DEL, FWD
                    svstart = curr.ref_start
                    svlen = (curr.ref_start - last.ref_end)
                    curr.svtypes_starts_lens.append(("DEL", svstart, -svlen))
                    hints += 1

                elif (rev and (last.ref_start - curr.ref_end) >= minsvlen_screen
                      and (last.ref_start - curr.ref_end) - (curr.qry_start - last.qry_end) >= minsvlen_screen):
                    # DEL, REV
                    svstart = last.ref_start
                    svlen = (last.ref_start - curr.ref_end)
                    curr.svtypes_starts_lens.append(("DEL", svstart, -svlen))
                    hints += 1

                elif fwd and curr.ref_start <= last.ref_end:
                    # DUP, FWD
                    svstart = curr.ref_start
                    svlen = (last.ref_end - curr.ref_start)
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("DUP", svstart, svlen))
                        hints += 1

                elif rev and last.ref_start <= curr.ref_end:
                    # DUP, REV
                    svstart = last.ref_start
                    svlen = (curr.ref_end - last.ref_start)
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("DUP", svstart, svlen))
                        hints += 1

            else:
                #
                # INV
                #
                if fwd and curr.ref_start <= last.ref_start:
                    # CASE B
                    svstart = curr.ref_start
                    svlen = last.ref_start - curr.ref_start
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("INV", svstart, svlen))
                        hints += 1

                elif fwd and curr.ref_start > last.ref_start:
                    # CASE C
                    svstart = last.ref_start
                    svlen = curr.ref_start - last.ref_start
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("INV", svstart, svlen))
                        hints += 1

                elif rev and curr.ref_end >= last.ref_end:
                    # CASE A
                    svstart = last.ref_end
                    svlen = curr.ref_end - last.ref_end
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("INV", svstart, svlen))
                        hints += 1

                elif rev and curr.ref_end < last.ref_end:
                    # CASE D
                    svstart = curr.ref_end
                    svlen = last.ref_end - curr.ref_end
                    if svlen >= minsvlen_screen:
                        curr.svtypes_starts_lens.append(("INV", svstart, svlen))
                        hints += 1
        else:
            #
            # BND
            #
            if curr.contig == main_contig:
                a, b = curr, last
            else:
                a, b = last, curr

            if a.contig == main_contig and abs(last.qry_end - last.qry_start) >= min_split_len_bnd and abs(curr.qry_end - curr.qry_start) >= min_split_len_bnd:
                is_first = a.qry_start < b.qry_start
                if is_first:
                    if a.strand == "+":
                        svstart = a.ref_end
                    else:
                        svstart = a.ref_start
                else:
                    if a.strand == "+":
                        svstart = a.ref_start
                    else:
                        svstart = a.ref_end
                a.svtypes_starts_lens.append(("BND",
                                              svstart,
                                              SVCallBNDInfo(b.contig,
                                                            b.ref_start,
                                                            is_first,
                                                            a.strand != b.strand)))
                hints += 1
        last = curr

    if not hints and len(leads) > 2:
        # filter out short pseudo BND hints
        left = leads[0]
        leads = [ld for ld in leads if ld.contig == left.contig and ld.strand == left.strand]
        if len(leads) == 2:
            return classify_splits(read, leads, config, main_contig)

    return leads


#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     02.01.2024
# Author:      Hermann Romanek
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
from typing import Tuple, Callable
from unittest import TestCase
from unittest.mock import Mock

from sniffles.sv import SVCall
from sniffles.vcf import VCF


class TestVCFFormat(TestCase):
    """
    Unittests for writing VCF output
    """

    @staticmethod
    def get_config():
        """
        Generate a mock config
        """
        config = Mock()
        config.sample_ids_vcf = []
        config.output_rnames = True
        config.mosaic_af_max = 0.3
        config.mosaic = False
        config.id_prefix = 'Sniffles.'
        config.symbolic = False
        config.max_del_seq_len = 50000
        config.genotype_format = "GT:GQ:DR:DV"
        return config

    def verify_common_fields(self, *args, **kwargs) -> Tuple[int, str, str, str]:
        """
        Verifies common fields, returns pos, ref and alt fields
        """
        flds = args[0].split('\t')
        self.assertGreater(len(flds), 8)
        contig, pos, id, ref, alt, qual, filter, info = flds[:8]

        self.assertEqual('chr1', contig)
        self.assertEqual('Sniffles.unittest-1', id)

        return int(pos), ref, alt, info

    def parse_info(self, info: str) -> tuple[set[str], dict[str, str]]:
        """
        Simple parsing of an info string
        """
        flags = []
        flds = {}
        for fld in info.split(';'):
            if '=' in fld:
                key, value = fld.split('=')
                flds[key] = value
            else:
                flags.append(fld)
        return set(flags), flds

    def get_vcf(self, reference: str):
        vcf = VCF(self.get_config(), None)
        vcf.reference_handle = Mock()
        vcf.reference_handle.fetch = Mock(
            side_effect=lambda refname, start, end: reference[start:end]
        )
        return vcf

    def get_svcall(self, **kwargs):
        sv_kwargs = {
            'contig': 'chr1',
            'id': 'unittest-1',
            'qual': 10,
            'filter': 'PASS',
            'info': {},
            'genotypes': {},
            'precise': True,
            'support': 100,
            'rnames': ['ut'],
            'postprocess': None,
            'qc': True,
            'nm': -1,
            'fwd': 1,
            'rev': 1,
        }
        sv_kwargs.update(kwargs)
        return SVCall(**sv_kwargs)

    def test_spec_ins(self):
        """
        Test from VCF spec 4.2, chapter 5.2.2

        #CHROM POS ID REF ALT   QUAL FILTER INFO
        20     3   .  C   CTAG  .    PASS   DP=100

        Ref    a t C - - - g a
        1      a t C T A G g a
        """
        def verify(*args, **kwargs):
            pos, ref, alt, _ = self.verify_common_fields(*args, **kwargs)
            self.assertEqual(3, pos)
            self.assertEqual('C', ref)
            self.assertEqual('CTAG', alt)

        vcf = self.get_vcf('atCga')
        vcf.write_raw = Mock(side_effect=verify)

        vcf.write_call(self.get_svcall(
            svtype='INS',
            ref='N',
            alt='TAG',
            pos=3,
            svlen=3,
            end=3,
        ))

        vcf.reference_handle.fetch.assert_called_with('chr1', 2, 3)
        vcf.write_raw.assert_called()

    def test_spec_del(self):
        """
        Test from VCF spec 4.2, chapter 5.2.3

        #CHROM POS ID REF ALT QUAL FILTER INFO
        20     2   .  TCG T   .    PASS   DP=100

        Ref  a T C G a
        Alt  a T - - a
        """
        def verify(*args, **kwargs):
            pos, ref, alt, info = self.verify_common_fields(*args, **kwargs)
            self.assertEqual(2, pos)
            self.assertEqual('TCG', ref)
            self.assertEqual('T', alt)
            _, info_fields = self.parse_info(info)
            self.assertEqual(pos + abs(int(info_fields['SVLEN'])) - 1, int(info_fields['END']))

        vcf = self.get_vcf('aTCGa')
        vcf.write_raw = Mock(side_effect=verify)

        vcf.write_call(self.get_svcall(
            svtype='DEL',
            ref='N',
            alt='<DEL>',
            pos=2,
            svlen=-2,
            end=4,
        ))

        vcf.write_raw.assert_called()

    def test_del_issue31(self):
        """
        See https://github.com/fritzsedlazeck/Sniffles_dev/issues/31
        """
        def mock_fetch(refname, start, end):
            """
            Reference is extracted from:
            reference_file = pysam.FastaFile('GRCh38_GIABv3_no_alt_analysis_set_maskedGRC_decoys_MAP2K3_KMT2C_KCNJ18.fasta.gz')
            reference_file.fetch('chr1', 964600, 964800)
            """
            reference = 'CAGTGGGGATGTGCTGCGGGGAGGGGGGCGCGGGTCCGCAGTGGGGATGTGCTGCCGGGAGGGGGGCGCGGGTCCGCAGTGGGGATGTGCTGCCGGGAGGGGGGCGCGGGTCCGCAGTGGGGATGTGCTGCCGGGAGGGGGGCGCGGGTCCGCAGTGGGGATGTGCTGCCGGGAGGGGGGCGCGGGTCCGCAGTGGGGAT'
            return reference[start-964600:end-964600]

        vcf = VCF(self.get_config(), None)
        vcf.reference_handle = Mock()
        vcf.reference_handle.fetch = Mock(
            side_effect=mock_fetch
        )

        def verify(*args, **kwargs):
            pos, ref, alt, info = self.verify_common_fields(*args, **kwargs)
            self.assertEqual(964631, pos)
            self.assertEqual('CGGGTCCGCAGTGGGGATGTGCTGCCGGGAGGGGGGCGCGGGTCCGCAGTGGGGATGTGCTGCCGGGAGGGGGGCG', ref)
            self.assertEqual('C', alt)

        vcf.write_raw = Mock(side_effect=verify)

        vcf.write_call(self.get_svcall(
            svtype='DEL',
            ref='N',
            alt='<DEL>',
            pos=964631,
            svlen=-75,
            end=964631-75,
        ))

        vcf.write_raw.assert_called()

    def test_unresolved_ins(self):
        """
        Tests output of an SV where we were unable to resolve the sequence.
          See https://github.com/fritzsedlazeck/Sniffles/issues/501
        """
        def verify(*args, **kwargs):
            pos, ref, alt, info = self.verify_common_fields(*args, **kwargs)
            self.assertEqual(2, pos)
            self.assertEqual('T', ref)
            self.assertEqual('<INS>', alt)

        vcf = self.get_vcf('T'*50)
        vcf.write_raw = Mock(side_effect=verify)

        vcf.write_call(self.get_svcall(
            svtype='INS',
            ref='N',
            alt='<INS>',
            pos=2,
            svlen=20,
            end=22,
        ))

        vcf.write_raw.assert_called()



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     28.05.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
from dataclasses import dataclass
import re
import itertools
from typing import Optional

import pysam

# for: --dev-cache
import os
import sys
import pickle
# end: for: --dev-cache

from sniffles import util
from sniffles import sv
from sniffles.region import Region


@dataclass
class Lead:
    read_id: int = None  # or tuple[int, str, str] for phased reads, with (read_id, HP, PS)
    read_qname: str = None
    contig: str = None
    ref_start: int = None
    ref_end: int = None
    qry_start: int = None
    qry_end: int = None
    strand: str = None
    mapq: int = None
    nm: float = None
    source: str = None
    svtype: str = None
    svlen: Optional[int] = None
    seq: Optional[str] = None
    svtypes_starts_lens: list = None


def CIGAR_analyze(cigar):
    buf = ""
    readspan = 0
    refspan = 0
    clip_start = None
    clip = 0
    for c in cigar:
        if c.isnumeric():
            buf += c
        else:
            oplen = int(buf)
            h = False
            if c in "MIX=":
                readspan += oplen
                h = True
            if c in "MDX=N":
                refspan += oplen
                h = True
            if not h:
                if c in "SH":
                    if clip_start == None and readspan + refspan > 0:
                        clip_start = clip
                    clip += oplen
                else:
                    raise f"Unknown CIGAR operation: '{c}'"
            buf = ""
    if clip_start == None:
        clip_start = clip
    return clip_start, clip - clip_start, refspan, readspan


def CIGAR_analyze_regex(cigar):
    # TODO: Obsolete
    opsums = {"M": 0, "I": 0, "D": 0, "=": 0, "X": 0, "N": 0, "H": 0, "S": 0}
    iter = re.split(r"(\d+)", cigar)
    for i in range(1, len(iter) - 1, 2):
        op = iter[i + 1]
        if op != "H" and op != "S":
            readstart_fwd = opsums["H"] + opsums["S"]
        opsums[iter[i + 1]] += int(iter[i])
    readstart_rev = opsums["H"] + opsums["S"] - readstart_fwd
    refspan = opsums["M"] + opsums["D"] + opsums["="] + opsums["X"] + opsums["N"]
    readspan = opsums["M"] + opsums["I"] + opsums["="] + opsums["X"]
    return readstart_fwd, readstart_rev, refspan, readspan


def CIGAR_tolist(cigar):
    # TODO: Obsolete (see CIGAR_tolist_analyze)
    """
    CIGAR string : str -> List of CIGAR operation & length tuples : [(op1:char, op1_length:int),...]
    """
    buf = ""
    ops = []
    for c in cigar:
        if c.isnumeric():
            buf += c
        else:
            ops.append((c, int(buf)))
            buf = ""
    return ops


def CIGAR_listrefspan(ops):
    # TODO: Obsolete (see CIGAR_analyze)
    # TODO(Potential): Detect&utilize minimap2 condensed supplementary alignment cigar for speed
    return sum(oplen for op, oplen in ops if op == "M" or op == "D" or op == "=" or op == "X" or op == "N")


def CIGAR_listreadspan(ops):
    # TODO: Obsolete (see CIGAR_analyze)
    return sum(oplen for op, oplen in ops if op == "M" or op == "I" or op == "=" or op == "X")


def CIGAR_listreadstart_fwd(ops):
    # TODO: Obsolete (see CIGAR_analyze)
    """
    Position in query read where CIGAR alignment starts (i.e. taking into account start clipping)
    """
    op, oplen = ops[0]
    op2, op2len = ops[1]
    if op == "H" or op == "S":
        assert (op2 != "H" and op2 != "S")
        return oplen
    else:
        return 0


def CIGAR_listreadstart_rev(ops):
    # TODO: Obsolete (see CIGAR_analyze)
    """
    Position in query read where CIGAR alignment starts (i.e. taking into account start clipping)
    """
    op, oplen = ops[-1]
    op2, op2len = ops[-2]
    if op == "H" or op == "S":
        assert (op2 != "H" and op2 != "S")
        return oplen
    else:
        return 0


OPTAB = {pysam.CMATCH: (1, 1, 0),
         pysam.CEQUAL: (1, 1, 0),
         pysam.CDIFF: (1, 1, 0),
         pysam.CINS: (1, 0, 1),
         pysam.CDEL: (0, 1, 1),
         pysam.CREF_SKIP: (0, 1, 0),
         pysam.CSOFT_CLIP: (1, 0, 1),
         pysam.CHARD_CLIP: (0, 0, 0),
         pysam.CPAD: (0, 0, 0)}
#      pysam.CBACK:      (0,0,0)}

OPLIST = [(0, 0, 0) for i in range(max(int(k) for k in OPTAB.keys()) + 1)]
for k, v in OPTAB.items():
    OPLIST[int(k)] = v


def read_iterindels(read_id, read, contig, config, use_clips, read_nm):
    minsvlen = config.minsvlen_screen
    longinslen = config.long_ins_length / 2.0
    seq_cache_maxlen = config.dev_seq_cache_maxlen
    qname = read.query_name
    mapq = read.mapping_quality
    strand = "-" if read.is_reverse else "+"
    CINS = pysam.CINS
    CDEL = pysam.CDEL
    CSOFT_CLIP = pysam.CSOFT_CLIP

    pos_read = 0
    pos_ref = read.reference_start

    for op, oplength in read.cigartuples:
        add_read, add_ref, event = OPLIST[op]
        if event and oplength >= minsvlen:
            if op == CINS:
                yield Lead(read_id,
                           qname,
                           contig,
                           pos_ref,
                           pos_ref,
                           pos_read,
                           pos_read + oplength,
                           strand,
                           mapq,
                           read_nm,
                           "INLINE",
                           "INS",
                           oplength,
                           seq=read.query_sequence[pos_read:pos_read + oplength] if oplength <= seq_cache_maxlen else None)
            elif op == CDEL:
                yield Lead(read_id,
                           qname,
                           contig,
                           pos_ref + oplength,
                           pos_ref,
                           pos_read,
                           pos_read,
                           strand,
                           mapq,
                           read_nm,
                           "INLINE",
                           "DEL",
                           -oplength)
            elif use_clips and op == CSOFT_CLIP and oplength >= longinslen:
                yield Lead(read_id,
                           qname,
                           contig,
                           pos_ref,
                           pos_ref,
                           pos_read,
                           pos_read + oplength,
                           strand,
                           mapq,
                           read_nm,
                           "INLINE",
                           "INS",
                           None,
                           seq=None)
        pos_read += add_read * oplength
        pos_ref += add_ref * oplength


def get_cigar_indels(read_id, read, contig, config, use_clips, read_nm):
    minsvlen = config.minsvlen_screen
    longinslen = config.long_ins_length / 2.0
    seq_cache_maxlen = config.dev_seq_cache_maxlen
    qname = read.query_name
    mapq = read.mapping_quality
    strand = "-" if read.is_reverse else "+"
    CINS = pysam.CINS
    CDEL = pysam.CDEL
    CSOFT_CLIP = pysam.CSOFT_CLIP

    INS_SUM = 0
    DEL_SUM = 0

    pos_read = 0
    pos_ref = read.reference_start
    for op, oplength in read.cigartuples:
        add_read, add_ref, event = OPLIST[op]
        if event:
            if op == CINS:
                INS_SUM += oplength
            elif op == CDEL:
                DEL_SUM += oplength
        pos_read += add_read * oplength
        pos_ref += add_ref * oplength

    return INS_SUM, DEL_SUM


def read_itersplits_bnd(read_id, read, contig, config, read_nm):
    assert (read.is_supplementary)
    # SA:refname,pos,strand,CIGAR,MAPQ,NM
    all_leads = []
    supps = [part.split(",") for part in read.get_tag("SA").split(";") if len(part) > 0]

    if len(supps) > config.max_splits_base + config.max_splits_kb * (read.query_length / 1000.0):
        return

    if read.is_reverse:
        qry_start = read.query_length - read.query_alignment_end
    else:
        qry_start = read.query_alignment_start

    curr_lead = Lead(read_id,
                     read.query_name,
                     contig,
                     read.reference_start,
                     read.reference_start + read.reference_length,
                     qry_start,
                     qry_start + read.query_alignment_length,
                     "-" if read.is_reverse else "+",
                     read.mapping_quality,
                     read_nm,
                     "SPLIT_SUP",
                     "?")
    all_leads.append(curr_lead)

    prim_refname, prim_pos, prim_strand, prim_cigar, prim_mapq, prim_nm = supps[0]
    if prim_refname == contig:
        # Primary alignment is on this chromosome, no need to parse the supplementary
        return

    minpos_curr_chr = min(itertools.chain([read.reference_start], (int(pos) for refname, pos, strand, cigar, mapq, nm in supps if refname == contig)))
    if minpos_curr_chr < read.reference_start:
        # Only process splits once per chr (there may be multiple supplementary alignments on the same chr)
        return

    for refname, pos, strand, cigar, mapq, nm in supps:
        mapq = int(mapq)
        nm = int(nm)
        # if not config.dev_keep_lowqual_splits and mapq < config.mapq:
        #    continue

        is_rev = (strand == "-")

        try:
            readstart_fwd, readstart_rev, refspan, readspan = CIGAR_analyze(cigar)
        except Exception as e:
            util.error(f"Malformed CIGAR '{cigar}' with pos {pos} of read '{read.query_name}' ({e}). Skipping.")
            return

        pos_zero = int(pos) - 1
        split_qry_start = readstart_rev if is_rev else readstart_fwd

        all_leads.append(Lead(read_id,
                              read.query_name,
                              refname,
                              pos_zero,
                              pos_zero + refspan,
                              split_qry_start,
                              split_qry_start + readspan,
                              strand,
                              mapq,
                              read_nm,
                              "SPLIT_SUP",
                              "?"))

    sv.classify_splits(read, all_leads, config, contig)

    for lead in all_leads:
        for svtype, svstart, arg in lead.svtypes_starts_lens:
            if svtype == "BND":
                bnd = Lead(read_id=lead.read_id,
                           read_qname=lead.read_qname,
                           contig=lead.contig,
                           ref_start=svstart,
                           ref_end=svstart,
                           qry_start=lead.qry_start,
                           qry_end=lead.qry_end,
                           strand=lead.strand,
                           mapq=lead.mapq,
                           nm=lead.nm,
                           source=lead.source,
                           svtype=svtype,
                           svlen=config.bnd_cluster_length,
                           seq=None)
                bnd.bnd_info = arg
                # print(lead.contig,svstart,bnd.bnd_info)
                yield bnd


def read_itersplits(read_id, read, contig, config, read_nm):
    # SA:refname,pos,strand,CIGAR,MAPQ,NM
    all_leads = []
    supps = [part.split(",") for part in read.get_tag("SA").split(";") if len(part) > 0]
    trace_read = config.dev_trace_read != False and config.dev_trace_read == read.query_name

    if len(supps) > config.max_splits_base + config.max_splits_kb * (read.query_length / 1000.0):
        return

    if trace_read:
        print(f"[DEV_TRACE_READ] [0c/4] [LeadProvider.read_itersplits] [{read.query_name}] passed max_splits check")

    # QC on: 18Aug21, HG002.ont.chr22; O.K.
    # cigarl=CIGAR_tolist(read.cigarstring)
    # if read.is_reverse:
    #    cigarl.reverse()

    # if read.is_reverse:
    #    assert(read.query_length-read.query_alignment_end == CIGAR_listreadstart(cigarl))
    # else:
    #    assert(read.query_alignment_start == CIGAR_listreadstart(cigarl))

    # assert(CIGAR_listrefspan(cigarl)==read.reference_length)
    # assert(CIGAR_listreadspan(cigarl)==read.query_alignment_length)
    # End QC

    if read.is_reverse:
        qry_start = read.query_length - read.query_alignment_end
    else:
        qry_start = read.query_alignment_start

    curr_lead = Lead(read_id,
                     read.query_name,
                     contig,
                     read.reference_start,
                     read.reference_start + read.reference_length,
                     qry_start,
                     qry_start + read.query_alignment_length,
                     "-" if read.is_reverse else "+",
                     read.mapping_quality,
                     read_nm,
                     "SPLIT_PRIM",
                     "?")
    all_leads.append(curr_lead)

    # QC on: 18Aug21; O.K.
    # assert(read.reference_length == CIGAR_listrefspan(CIGAR_tolist(read.cigarstring)))
    # assert(read.query_alignment_start == CIGAR_listreadstart(CIGAR_tolist(read.cigarstring)))
    # assert(read.query_alignment_length == CIGAR_listreadspan(CIGAR_tolist(read.cigarstring)))
    # End QC

    for refname, pos, strand, cigar, mapq, nm in supps:
        mapq = int(mapq)
        nm = int(nm)
        # if not config.dev_keep_lowqual_splits and mapq < config.mapq:
        #    continue

        is_rev = (strand == "-")

        try:
            readstart_fwd, readstart_rev, refspan, readspan = CIGAR_analyze(cigar)
        except Exception as e:
            util.error(f"Malformed CIGAR '{cigar}' with pos {pos} of read '{read.query_name}' ({e}). Skipping.")
            return

        pos_zero = int(pos) - 1
        split_qry_start = readstart_rev if is_rev else readstart_fwd

        all_leads.append(Lead(read_id,
                              read.query_name,
                              refname,
                              pos_zero,
                              pos_zero + refspan,
                              split_qry_start,
                              split_qry_start + readspan,
                              strand,
                              mapq,
                              read_nm,
                              "SPLIT_SUP",
                              "?"))

        # QC on: 08Sep21; O.K.
        # cigarl=CIGAR_tolist(cigar)
        # assert(CIGAR_listrefspan(cigarl)==refspan)
        # assert(CIGAR_listreadspan(cigarl)==readspan)
        # assert(CIGAR_listreadstart_fwd(cigarl)==readstart_fwd)
        # assert(CIGAR_listreadstart_rev(cigarl)==readstart_rev)
        # End QC

    if trace_read:
        print(f"[DEV_TRACE_READ] [0c/4] [LeadProvider.read_itersplits] [{read.query_name}] all_leads: {all_leads}")

    all_leads = sv.classify_splits(read, all_leads, config, contig)

    if trace_read:
        print(f"[DEV_TRACE_READ] [0c/4] [LeadProvider.read_itersplits] [{read.query_name}] classify_splits(all_leads): {all_leads}")

    """
    if config.dev_trace_read != False:
        print(read.query_name)
        if read.query_name == config.dev_trace_read:
            for lead_i, lead in enumerate(all_leads):
                for svtype, svstart, arg in lead.svtypes_starts_lens:
                    min_mapq=min(lead.mapq,all_leads[max(0,lead_i-1)].mapq)
                    keep=True
                    if not config.dev_keep_lowqual_splits and min_mapq < config.mapq:
                        keep=False
                    print(f"[DEV_TRACE_READ] [REPORT_LEAD_SPLIT] Splits identified from read {read.read_id}")
    """

    for lead_i, lead in enumerate(all_leads):
        for svtype, svstart, arg in lead.svtypes_starts_lens:
            min_mapq = min(lead.mapq, all_leads[max(0, lead_i - 1)].mapq)
            if not config.dev_keep_lowqual_splits and min_mapq < config.mapq:
                continue

            if svtype == "BND":
                bnd = Lead(read_id=lead.read_id,
                           read_qname=lead.read_qname,
                           contig=lead.contig,
                           ref_start=svstart,
                           ref_end=svstart,
                           qry_start=lead.qry_start,
                           qry_end=lead.qry_end,
                           strand=lead.strand,
                           mapq=lead.mapq,
                           nm=lead.nm,
                           source=lead.source,
                           svtype=svtype,
                           svlen=config.bnd_cluster_length,
                           seq=None)
                bnd.bnd_info = arg
                yield bnd

            elif svtype != "NOSV":
                svlen = arg

                yield Lead(read_id=lead.read_id,
                           read_qname=lead.read_qname,
                           contig=lead.contig,
                           ref_start=svstart,
                           ref_end=svstart + svlen if svlen != None and svtype != "INS" else svstart,
                           qry_start=lead.qry_start,
                           qry_end=lead.qry_end,
                           strand=lead.strand,
                           mapq=lead.mapq,
                           nm=lead.nm,
                           source=lead.source,
                           svtype=svtype,
                           svlen=svlen,
                           seq=lead.seq if svtype == "INS" else None)


class LeadProvider:
    def __init__(self, config, read_id_offset):
        self.config = config

        self.leadtab = {}
        self.leadcounts = {}

        for svtype in sv.TYPES:
            self.leadtab[svtype] = {}
            self.leadcounts[svtype] = 0

        self.covrtab_fwd = {}
        self.covrtab_rev = {}
        self.covrtab_min_bin = None
        # self.covrtab_read_start={}
        # self.covrtab_read_end={}

        self.read_id = read_id_offset
        self.read_count = 0

        self.contig = None
        self.start = None
        self.end = None

    def record_lead(self, ld, pos_leadtab):
        leadtab_svtype = self.leadtab[ld.svtype]
        if pos_leadtab in leadtab_svtype:
            leadtab_svtype[pos_leadtab].append(ld)
            lead_count = len(leadtab_svtype[pos_leadtab])
            if lead_count > self.config.consensus_max_reads_bin:
                ld.seq = None
        else:
            leadtab_svtype[pos_leadtab] = [ld]
            lead_count = 1
        self.leadcounts[ld.svtype] += 1

    def build_leadtab(self, regions: list[Region], bam):

        assert (self.contig is None)
        assert (self.start is None)
        assert (self.end is None)

        externals = []
        ld_binsize = self.config.cluster_binsize

        for region in regions:
            self.contig = region.contig
            self.start = region.start if self.start is None else min(region.start, self.start)
            self.end = region.end if self.end is None else max(region.start, self.end)
            self.covrtab_min_bin = int(self.start / self.config.coverage_binsize) * self.config.coverage_binsize

            for ld in self.iter_region(bam, region):
                ld_contig, ld_ref_start = ld.contig, ld.ref_start

                if region.contig == ld_contig and region.start <= ld_ref_start < region.end:
                    pos_leadtab = int(ld_ref_start / ld_binsize) * ld_binsize
                    self.record_lead(ld, pos_leadtab)
                else:
                    externals.append(ld)

        return externals

    def iter_region(self, bam, region: Region):
        leads_all = []
        binsize = self.config.cluster_binsize
        coverage_binsize = self.config.coverage_binsize
        coverage_shift_bins = self.config.coverage_shift_bins
        coverage_shift_min_aln_len = self.config.coverage_shift_bins_min_aln_length
        long_ins_threshold = self.config.long_ins_length * 0.5
        qc_nm = self.config.qc_nm_measure
        phase = self.config.phase
        advanced_tags = qc_nm or phase
        mapq_min = self.config.mapq
        alen_min = self.config.min_alignment_length
        nm_sum = 0
        nm_count = 0
        trace_read = self.config.dev_trace_read

        read_mq20 = 0
        for read in bam.fetch(region.contig, region.start, region.end, until_eof=False):
            if trace_read is not False:
                if trace_read == read.query_name:
                    print(f"[DEV_TRACE_READ] [0b/4] [LeadProvider.iter_region] [{region}] [{read.query_name}] has been fetched and is entering pre-filtering")

            # if self.read_count % 1000000 == 0:
            #    gc.collect()
            if read.reference_start < region.start or read.reference_start >= region.end:
                continue

            self.read_id += 1
            self.read_count += 1

            alen = read.query_alignment_length
            if read.mapping_quality < mapq_min or read.is_secondary or alen < alen_min:
                continue

            has_sa = read.has_tag("SA")
            use_clips = self.config.detect_large_ins and not read.is_supplementary and not has_sa

            nm = -1
            curr_read_id = self.read_id
            if advanced_tags:
                if qc_nm:
                    if read.has_tag("NM"):
                        nm_raw = read.get_tag("NM")
                        nm_ratio = read.get_tag("NM") / float(read.query_alignment_length + 1)
                        ins_sum, del_sum = get_cigar_indels(curr_read_id, read, region.contig, self.config, use_clips, read_nm=nm)
                        nm_adj = (nm_raw - (ins_sum + del_sum))
                        nm_adj_ratio = nm_adj / float(read.query_alignment_length + 1)
                        nm = nm_adj_ratio
                        nm_sum += nm
                        nm_count += 1

                if phase:
                    curr_read_id = (self.read_id, str(read.get_tag("HP")) if read.has_tag("HP") else "NULL", str(read.get_tag("PS")) if read.has_tag("PS") else "NULL")

            if trace_read is not False:
                if trace_read == read.query_name:
                    print(f"[DEV_TRACE_READ] [0b/4] [LeadProvider.iter_region] [{region}] [{read.query_name}] passed pre-filtering (whole-read), begin to extract leads")

            # Extract small indels
            for lead in read_iterindels(curr_read_id, read, region.contig, self.config, use_clips, read_nm=nm):
                if trace_read is not False:
                    if trace_read == read.query_name:
                        print(f"[DEV_TRACE_READ] [1/4] [leadprov.read_iterindels] [{region}] [{read.query_name}] new lead: {lead}")
                yield lead

            # Extract read splits
            if has_sa:
                if read.is_supplementary:
                    if trace_read is not False:
                        if trace_read == read.query_name:
                            print(f"[DEV_TRACE_READ] [1/4] [leadprov.read_itersplits_bnd] [{region}] [{read.query_name}] is entering read_itersplits_bnd")
                    for lead in read_itersplits_bnd(curr_read_id, read, region.contig, self.config, read_nm=nm):
                        if trace_read is not False:
                            if trace_read == read.query_name:
                                print(f"[DEV_TRACE_READ] [1/4] [leadprov.read_itersplits_bnd] [{region}] [{read.query_name}] new lead: {lead}")
                        yield lead
                else:
                    if trace_read is not False:
                        if trace_read == read.query_name:
                            print(f"[DEV_TRACE_READ] [1/4] [leadprov.read_itersplits] [{region}] [{read.query_name}] is entering read_itersplits")
                    for lead in read_itersplits(curr_read_id, read, region.contig, self.config, read_nm=nm):
                        if trace_read is not False:
                            if trace_read == read.query_name:
                                print(f"[DEV_TRACE_READ] [1/4] [leadprov.read_itersplits] [{region}] [{read.query_name}] new lead: {lead}")
                        yield lead
                read_mq20 += 1 if read.mapping_quality >= 20 else 0

            # Record in coverage table
            read_end = read.reference_start + read.reference_length
            assert (read_end == read.reference_end)
            # assert(read_end>=read.reference_start)
            if read.is_reverse:
                target_tab = self.covrtab_rev
            else:
                target_tab = self.covrtab_fwd
            covr_start_bin = (int(read.reference_start / coverage_binsize) + coverage_shift_bins * (alen >= coverage_shift_min_aln_len)) * coverage_binsize
            covr_end_bin = (int(read_end / coverage_binsize) - coverage_shift_bins * (alen >= coverage_shift_min_aln_len)) * coverage_binsize

            if covr_end_bin > covr_start_bin:
                self.covrtab_min_bin = min(self.covrtab_min_bin, covr_start_bin)
                target_tab[covr_start_bin] = target_tab[covr_start_bin] + 1 if covr_start_bin in target_tab else 1

                if read_end <= self.end:
                    target_tab[covr_end_bin] = target_tab[covr_end_bin] - 1 if covr_end_bin in target_tab else -1

        self.config.average_regional_nm = nm_sum / float(max(1, nm_count))
        self.config.qc_nm_threshold = self.config.average_regional_nm
        # print(f"Contig {contig} avg. regional NM={self.config.average_regional_nm}, threshold={self.config.qc_nm_threshold}")

    def dev_leadtab_filename(self, contig, start, end):
        scriptloc = os.path.dirname(os.path.realpath(sys.argv[0]))
        if self.config.dev_cache_dir is None:
            cache_dir = f"{scriptloc}/cache"
        else:
            cache_dir = self.config.dev_cache_dir
        return f"{cache_dir}/{os.path.basename(self.config.input)}_{contig}_{start}_{end}.pickle"



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     01.07.2024
# Author:      Hermann Romanek
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
import logging
from dataclasses import dataclass
from typing import Optional

log = logging.getLogger(__name__)


@dataclass
class Region:
    """
    A region from a bed file.

    Coordinates are 0-based, half open
    """
    contig: str
    start: int
    end: int

    @classmethod
    def from_bed_line(cls, line: str) -> Optional['Region']:
        try:
            contig, start, end = line.split('\t')[:3]
            return Region(
                contig=contig,
                start=int(start),
                end=int(end),
            )
        except ValueError as ex:
            log.warning(f'Invalid region line: {line} ({ex})')
            return None

    def __str__(self) -> str:
        return f'{self.contig}:{self.start}-{self.end}'
